<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace AutoreportModule\Command;


use App\Lib\Exceptions\MissingDbBackendException;
use App\Model\Table\SystemsettingsTable;
use AutoreportModule\Lib\Autoreport;
use AutoreportModule\Lib\CsvReport;
use AutoreportModule\Model\Table\AutoreportQuarterlyTable;
use AutoreportModule\Model\Table\AutoreportSettingsTable;
use AutoreportModule\Model\Table\AutoreportsTable;
use Cake\Console\Arguments;
use Cake\Console\Command;
use Cake\Console\ConsoleIo;
use Cake\Console\ConsoleOptionParser;
use Cake\Filesystem\Folder;
use Cake\Mailer\Mailer;
use Cake\ORM\TableRegistry;
use itnovum\openITCOCKPIT\Core\Interfaces\CronjobInterface;
use itnovum\openITCOCKPIT\Core\Reports\CsvCreator;
use itnovum\openITCOCKPIT\Core\Reports\DaterangesCreator;
use itnovum\openITCOCKPIT\Core\Views\Logo;
use itnovum\openITCOCKPIT\Core\Views\UserTime;
use PuppeteerPdf\Pdf\PdfRenderer;

/**
 * Class AutoreportCommand
 * @package AutoreportModule\Command
 */
class AutoreportCommand extends Command implements CronjobInterface {

    /**
     * @var ConsoleIo
     */
    private $io;

    /**
     * Hook method for defining this command's option parser.
     *
     * @see https://book.cakephp.org/3.0/en/console-and-shells/commands.html#defining-arguments-and-options
     *
     * @param \Cake\Console\ConsoleOptionParser $parser The parser to be defined
     * @return \Cake\Console\ConsoleOptionParser The built parser.
     */
    public function buildOptionParser(ConsoleOptionParser $parser): ConsoleOptionParser {
        $parser = parent::buildOptionParser($parser);

        return $parser;
    }

    /**
     * @param Arguments $args
     * @param ConsoleIo $io
     * @return int|void|null
     */
    public function execute(Arguments $args, ConsoleIo $io) {
        $this->io = $io;
        $this->io->info(__('Delete temporary automatic report files...'));
        $this->clenupAutoreportCharts();
        $this->sendAutoreports();
        $this->cleanUp();
        $this->generateQuarterlyReports();
    }

    public function sendAutoreports() {

        /** @var  $SystemsettingsTable SystemsettingsTable */
        $SystemsettingsTable = TableRegistry::getTableLocator()->get('Systemsettings');
        $_systemsettings = $SystemsettingsTable->findAsArray();
        $Logo = new Logo();

        $this->io->setStyle('green', ['text' => 'green', 'blink' => false]);
        $this->io->setStyle('red', ['text' => 'red', 'blink' => false]);

        $this->io->out('Sending automatic reports ...');

        $toSend = false;
        /** @var $AutoreportsTable AutoreportsTable */
        $AutoreportsTable = TableRegistry::getTableLocator()->get('AutoreportModule.Autoreports');
        $autoreports = $AutoreportsTable->getAutoreportsForCronjob();

        /** @var AutoreportSettingsTable $AutoreportSettingsTable */
        $AutoreportSettingsTable = TableRegistry::getTableLocator()->get('AutoreportModule.AutoreportSettings');
        $settings = $AutoreportSettingsTable->currentAutoreportSettings();
        $defaultSettings = $AutoreportSettingsTable->defaultAutoreportSettings();

        $pdfSend = (boolean)($settings['configuration_option'] &
            $defaultSettings['AUTOREPORTS_SEND_REPORT_AS_PDF']['value']);
        $csvSend = (boolean)($settings['configuration_option'] &
            $defaultSettings['AUTOREPORTS_SEND_REPORT_AS_CSV']['value']);

        if (empty($settings['store_path']) && !$pdfSend && !$csvSend) {
            $this->io->out(__(' no reports to send'), 0);
            $this->io->success('    Ok');
            return;
        }

        foreach ($autoreports as $autoreport) {
            $nothingToDo = true;
            $subject = ($settings['configuration_option'] &
                $defaultSettings['AUTOREPORTS_TENANT_AS_SUBJECT']['value']) ? $autoreport['container']['name'] : $autoreport['name'];
            /** @var DaterangesCreator $has_to_be_send */
            try {
                $hasToBeSend = DaterangesCreator::hasToBeSend($autoreport['last_send_date'], $autoreport['report_send_interval']);
                if ($hasToBeSend === false) {
                    continue;
                }
                $emailsToSend = [];
                foreach ($autoreport['users'] as $user) {
                    $emailsToSend[] = $user->get('email');
                }
                if (empty($emailsToSend) && empty($settings['store_path'])) {
                    continue;
                }
                $reportStart = DaterangesCreator::getStartTime($autoreport['report_send_interval']);
                $reportEnd = DaterangesCreator::getEndTime($autoreport['report_send_interval']);

                $UserTime = new UserTime(date_default_timezone_get(), 'd.m.Y H:i:s');

                if ($autoreport['use_start_time'] && $autoreport['report_start_date'] != '0000-00-00') {
                    $startDate = $autoreport['report_start_date']->getTimestamp();
                    if ($startDate < $reportEnd) {
                        $reportStart = $startDate;
                    }
                }

                $this->io->out($autoreport['name']);
                $this->io->hr();
                $baseFilename = $autoreport['name'];
                $baseFilenameSuffix = '';
                if ($settings['configuration_option'] & $defaultSettings['AUTOREPORTS_USE_CREATE_DATE_IN_FILENAME']['value']) {
                    $baseFilenameSuffix .= '_' . date('dmY', time());
                }
                if ($settings['configuration_option'] & $defaultSettings['AUTOREPORTS_USE_YEAR_IN_FILENAME']['value']) {
                    if (strlen($baseFilenameSuffix) === 0) {
                        $baseFilenameSuffix .= '_' . date('Y', $reportStart);
                    }
                }
                $baseFilename .= $baseFilenameSuffix;

                $pdfFileName = $baseFilename . '.pdf';

                $Autoreport = new Autoreport($autoreport['id'], $reportStart, $reportEnd, $UserTime);
                $result = $Autoreport->canReportBeCreated();
                if (!empty($result['error'])) {
                    if (!empty($result['objects'])) {
                        $this->io->error(
                            __('An error occured while sending auto report mail: "{0}":',
                                $result['message']
                            ));
                        foreach ($result['objects'] as $object) {
                            $this->io->out("\t" . __('Host {0}', $object['name']));
                            if (!empty($object['services'])) {
                                foreach ($object['services'] as $service) {
                                    $this->io->out("\t\t" . __('Service {0}', $service));
                                }
                            }
                        }
                    } else {
                        $this->io->error(
                            __('An error occured while sending auto report mail: "{0}":',
                                $result['message']
                            )
                        );
                    }
                    continue;
                }

                $finalReport = $Autoreport->generateReport(true);

                $PdfRenderer = new PdfRenderer();
                $PdfRenderer->templatePath('Autoreports/pdf');
                $PdfRenderer->template('AutoreportModule.generate', 'default');
                $PdfRenderer->viewVars([
                    'report'      => $finalReport,
                    'reportStart' => $UserTime->format($reportStart),
                    'reportEnd'   => $UserTime->format($reportEnd),
                    'UserTime'    => $UserTime
                ]);
                // Get the PDF string returned
                $pdf = $PdfRenderer->output();

                if (!empty($settings['store_path'])) {
                    $savePath = $settings['store_path'] . DS;
                    if (!is_dir($savePath)) {
                        $tmp_dir = new Folder($savePath);
                    }
                    if ($pdfSend) {
                        if (!is_dir(($savePath))) {
                            mkdir($savePath);
                        }

                        $pdfFile = fopen($savePath . $pdfFileName, 'w+');
                        fwrite($pdfFile, $pdf);
                        fclose($pdfFile);
                        $nothingToDo = false;
                    }
                    if ($csvSend) {
                        $CsvReport = new CsvReport($finalReport, $reportStart, $savePath, $settings, $defaultSettings);
                        $csvZipFile = $CsvReport->getReportAsZip();
                        $nothingToDo = false;
                    }
                }

                if (!empty($autoreport['users'])) {
                    try {
                        foreach ($autoreport['users'] as $user) {
                            $Mailer = new Mailer();
                            $Mailer->setSubject($subject);
                            $Mailer->setFrom($_systemsettings['MONITORING']['MONITORING.FROM_ADDRESS'], $_systemsettings['MONITORING']['MONITORING.FROM_NAME']);
                            $Mailer->setEmailFormat('both');
                            $Mailer->setTo($user->get('email'));

                            $attachments = [
                                'logo.png' => [
                                    'file'      => $Logo->getSmallLogoDiskPath(),
                                    'mimetype'  => 'image/png',
                                    'contentId' => '100'
                                ],
                            ];

                            if ($pdfSend) {
                                $attachments[$pdfFileName] = [
                                    'data'     => $pdf,
                                    'mimetype' => 'application/pdf'
                                ];
                            }

                            if ($csvSend) {
                                $CsvReport = new CsvReport($finalReport, $reportStart, '/tmp/', $settings, $defaultSettings);
                                $csvZipFile = $CsvReport->getReportAsZip();
                                $basename = basename($csvZipFile);
                                $attachments[$basename] = [
                                    'file'     => $csvZipFile,
                                    'mimetype' => 'application/zip'
                                ];
                            }

                            $Mailer->setAttachments($attachments);
                            $Mailer->viewBuilder()
                                ->setTemplate('AutoreportModule.autoreport_mail')
                                ->setVar('autoReportName', $autoreport['name']);

                            $Mailer->deliver();
                            if ($csvSend) {
                                unlink($csvZipFile);
                            }
                            $this->io->info(__('Sending Report {0} to {1}', $autoreport['name'], $user->get('email')));
                        }
                        $toSend = true;
                        $this->io->success('Report with ID "' . $autoreport['id'] . '" sent to mail address "' . implode(', ', $emailsToSend) . '"');
                        $nothingToDo = false;
                    } catch (\Exception $ex) {
                        $this->io->error(__('An error occured while sending auto report mail: ') . $ex->getMessage());
                    }
                }
            } catch (\Exception $e) {
                $this->io->error($e->getMessage());
            }

            if (!$toSend) {
                $this->io->success('No emails to send');
            }
            if (!$nothingToDo) {
                $ReportToUpdate = $AutoreportsTable->get($autoreport['id']);
                if (!empty($finalReport['autoreport_data'])) {
                    $ReportToUpdate->set('last_number_of_outages', $finalReport['autoreport_data']['last_number_of_outages']);
                    $ReportToUpdate->set('last_percent_value',
                        round($finalReport['autoreport_data']['total_outage_time'] / $finalReport['autoreport_data']['total_time'], 3)
                    );
                    $ReportToUpdate->set('last_absolut_value', $finalReport['autoreport_data']['total_outage_time']);
                    $ReportToUpdate->set('failure_statistic', (
                        $finalReport['autoreport_data']['last_number_of_outages'] > $autoreport['last_number_of_outages']
                    ) ? 1 : 0);
                    $ReportToUpdate->set('last_send_date', date('o-m-d H:i:s', time()));

                    $AutoreportsTable->save($ReportToUpdate);
                    if ($ReportToUpdate->hasErrors()) {
                        $this->io->error(
                            __('Report "{0}" with ID: {1} run failed: {2}',
                                $autoreport['name'],
                                $autoreport['id'],
                                $ReportToUpdate->getErrors())
                        );
                    } else {
                        $this->io->success(
                            __('Report "{0}" with ID: {1} run successfully',
                                $autoreport['name'],
                                $autoreport['id'])
                        );
                    }
                }
            }
            $this->cleanUp();
            $this->io->hr();
        }
    }

    public function generateQuarterlyReports() {
        /** @var SystemsettingsTable $SystemsettingsTable */
        $SystemsettingsTable = TableRegistry::getTableLocator()->get('Systemsettings');
        $_systemsettings = $SystemsettingsTable->findAsArray();

        /** @var AutoreportSettingsTable $AutoreportSettingsTable */
        $AutoreportSettingsTable = TableRegistry::getTableLocator()->get('AutoreportModule.AutoreportSettings');
        $settings = $AutoreportSettingsTable->currentAutoreportSettings();
        $defaultSettings = $AutoreportSettingsTable->defaultAutoreportSettings();

        $quarterlySend = ($settings['configuration_option'] &
            $defaultSettings['AUTOREPORTS_GENERATE_QUARTERLY_REPORTS']['value']);

        $nothingToDo = true;
        if ($quarterlySend) {
            $UserTime = new UserTime(date_default_timezone_get(), 'd.m.Y H:i:s');
            $this->io->out('Generating automatic quarterly reports ...');
            $this->io->hr();
            $Logo = new Logo();

            if (empty($settings['store_path'])) {
                $this->io->error('Quarterly auto reports could not be generated! No store path defined!');
                if (!empty($settings['notification_emails'])) {
                    $this->io->info('Sending Error emails to ' . $settings['notification_emails'] . ' ...');
                    $Mailer = new Mailer();
                    $Mailer->setSubject(__('Error') . ' - ' . __('Cannot generate quarterly autoreports.'));
                    $Mailer->setFrom($_systemsettings['MONITORING']['MONITORING.FROM_ADDRESS'], $_systemsettings['MONITORING']['MONITORING.FROM_NAME']);
                    $Mailer->setEmailFormat('both');
                    $Mailer->setTo(explode(';', $settings['notification_emails']));
                    $Mailer->setAttachments([
                        'logo.png' => [
                            'file'      => $Logo->getSmallLogoDiskPath(),
                            'mimetype'  => 'image/png',
                            'contentId' => '100'
                        ]
                    ]);
                    $Mailer->viewBuilder()
                        ->setTemplate('AutoreportModule.default_mail')
                        ->setVar('quarterlyAutoreportmessage', __('Autoreports could not be generated! No store path defined!'));

                    if ($Mailer->deliver()) {
                        $this->io->success('Error email was successfully sent');
                    } else {
                        $this->io->error('Error email could not be sent!');
                    }
                }
                return false;
            }

            $wasGenerated = [];
            for ($quartLastYear = 1; $quartLastYear <= 4; $quartLastYear++) {
                $wasGenerated[intval(date('Y')) - 1][$quartLastYear]['start'] = DaterangesCreator::getQuarterDates(intval(date('Y')) - 1, $quartLastYear);
                $wasGenerated[intval(date('Y')) - 1][$quartLastYear]['end'] = DaterangesCreator::getQuarterDates(intval(date('Y')) - 1, $quartLastYear, false);
            }
            for ($quarterThisYear = 1; $quarterThisYear < intval(ceil(date('m') / 3)); $quarterThisYear++) {
                $wasGenerated[intval(date('Y'))][$quarterThisYear]['start'] = DaterangesCreator::getQuarterDates(intval(date('Y')), $quarterThisYear);
                $wasGenerated[intval(date('Y'))][$quarterThisYear]['end'] = DaterangesCreator::getQuarterDates(intval(date('Y')), $quarterThisYear, false);
            }

            /** @var AutoreportQuarterlyTable $AutoreportQuarterlyTable */
            $AutoreportQuarterlyTable = TableRegistry::getTableLocator()->get('AutoreportModule.AutoreportQuarterly');
            $allQuarterlyAutoreports = $AutoreportQuarterlyTable->getActuallyQuarterlies();

            /** @var $AutoreportsTable AutoreportsTable */
            $AutoreportsTable = TableRegistry::getTableLocator()->get('AutoreportModule.Autoreports');
            $autoreports = $AutoreportsTable->getAutoreportsForCronjob();

            $savePath = $settings['store_path'] . DS;
            if (!is_dir($savePath)) {
                $tmp_dir = new Folder($savePath);
            }
            foreach ($wasGenerated as $year => $quartArr) {
                foreach ($quartArr as $quarter => $datesArr) {
                    $quarterReports = [];
                    foreach ($autoreports as $autoreport) {
                        $alreadyGenerated = false;
                        foreach ($allQuarterlyAutoreports as $quarterlyAutoreport) {
                            if ($year === intval($quarterlyAutoreport['year']) &&
                                $quarter === intval($quarterlyAutoreport['quarter'])
                            ) {
                                $alreadyGenerated = true;
                                break;
                            }
                        }

                        if ($alreadyGenerated) {
                            continue;
                        }
                        $Autoreport = new Autoreport($autoreport['id'], $datesArr['start'], $datesArr['end'], $UserTime);
                        $result = $Autoreport->canReportBeCreated();
                        if (!empty($result['error'])) {
                            if (!empty($result['objects'])) {
                                $this->io->error(
                                    __('An error occured while generating quarterly autoreports: "{0}":',
                                        $result['message']
                                    ));
                                foreach ($result['objects'] as $object) {
                                    $this->io->out("\t" . __('Host {0}', $object['name']));
                                    if (!empty($object['services'])) {
                                        foreach ($object['services'] as $service) {
                                            $this->io->out("\t\t" . __('Service {0}', $service));
                                        }
                                    }
                                }
                            } else {
                                $this->io->error(
                                    __('An error occured while generating quarterly autoreports: "{0}":',
                                        $result['message']
                                    )
                                );
                            }

                            $Mailer = new Mailer();
                            $Mailer->setSubject(__('Error') . ' - ' . __('Cannot generate quarterly autoreports.'));
                            $Mailer->setFrom($_systemsettings['MONITORING']['MONITORING.FROM_ADDRESS'], $_systemsettings['MONITORING']['MONITORING.FROM_NAME']);
                            $Mailer->setEmailFormat('both');
                            $Mailer->setTo(explode(';', $settings['notification_emails']));
                            $Mailer->setAttachments([
                                'logo.png' => [
                                    'file'      => $Logo->getSmallLogoDiskPath(),
                                    'mimetype'  => 'image/png',
                                    'contentId' => '100'
                                ]
                            ]);
                            $Mailer->viewBuilder()
                                ->setTemplate('AutoreportModule.default_mail')
                                ->setVar('quarterlyAutoreportmessage', $result['message'])
                                ->setVar('quarterlyAutoreportObjects', !empty($result['objects']) ? $result['objects'] : []);

                            if ($Mailer->deliver()) {
                                $this->io->success('Error email was successfully sent');
                            } else {
                                $this->io->error('Error email could not be sent!');
                            }
                            continue;
                        }
                        try {
                            $quarterReport = $Autoreport->generateReport(false);
                        } catch (MissingDbBackendException $e) {
                            $this->io->error($e->getMessage());
                        }
                        if (!empty($quarterReport['autoreport_data'])) {
                            $quarterReports[$autoreport['id']] = $quarterReport;
                        }
                    }
                    if (!empty($quarterReports)) {
                        $filename = 'Report_Q_' . $quarter . '_' . $year . '.csv';
                        $this->io->out('Generating ' . $filename);
                        $fullpath = $savePath . $filename;
                        $Csv = new CsvCreator();
                        $Csv->setPath($fullpath);
                        $Csv->addRow([
                            __('Report id'),
                            __('Report name'),
                            __('Time periods'),
                            __('Reflection state'),
                            __('Downtimes'),
                            __('Minimal allowed availability (min)'),
                            __('From'),
                            __('To'),
                            __('Minimal allowed availability (%)'),
                            __('Current availability'),
                        ]);
                        foreach ($quarterReports as $quarterReport) {
                            $totalTimeInSec = $quarterReport['autoreport_data']['total_time'];
                            $totalOutageInSec = $quarterReport['autoreport_data']['total_outage_time'];
                            $weekDaysShort = $this->weekDaysShort($quarterReport['autoreport']['timeperiod']['timeperiod_timeranges']);
                            if ($autoreport['min_availability_percent']) {
                                $minAvailabilityInPercent = floatval($autoreport['min_availability_percent']);
                                $minAvailabilityInMin = round($totalTimeInSec * $minAvailabilityInPercent / 6000);
                            } else {
                                $minAvailabilityInMin = $autoreport['min_availability'];
                                $minAvailabilityInPercent = round($minAvailabilityInMin * 6000 / $totalTimeInSec, 3);
                            }
                            $availabilityInPercent = strval(round(100 * ($totalTimeInSec - $totalOutageInSec) / $totalTimeInSec, 3));
                            $Csv->addRow([
                                $quarterReport['autoreport']['id'],
                                $quarterReport['autoreport']['name'],
                                $weekDaysShort,
                                intval($quarterReport['autoreport']['check_hard_state']),
                                intval($quarterReport['autoreport']['consider_downtimes']),
                                $minAvailabilityInMin,
                                date('d.m.Y', $datesArr['start']),
                                date('d.m.Y', $datesArr['end']),
                                sprintf("%01.3f", $minAvailabilityInPercent),
                                sprintf("%01.3f", $availabilityInPercent)
                            ]);
                        }
                        $Csv->closeFile();

                        $quarterlyAutoreport = $AutoreportQuarterlyTable->newEmptyEntity();
                        $quarterlyAutoreport->set('year', $year);
                        $quarterlyAutoreport->set('quarter', $quarter);
                        $quarterlyAutoreport->set('filename', $filename);

                        $AutoreportQuarterlyTable->save($quarterlyAutoreport);
                        if ($quarterlyAutoreport->hasErrors()) {
                            $this->io->error(__('{0} was not generated.', $filename));
                        } else {
                            $this->io->out('Generating ' . $filename, 1);
                            $this->io->success(' Ok');
                        }
                        $nothingToDo = false;
                    }
                }
            }
        }
        if ($nothingToDo) {
            $this->io->success('There\'s nothing to do');
        }
    }

    public function clenupAutoreportCharts() {
        $path = ROOT . DS . 'plugins' . DS . 'AutoreportModule' . DS . 'webroot' . DS . 'tmp';
        if (is_dir($path)) {
            $folder = new Folder($path);
            foreach ($folder->find() as $file) {
                if (filemtime($path . DS . $file) < (time() - 120)) {
                    unlink($path . DS . $file);
                }
            }
        }
    }

    public function cleanUp() {
        $savePath = ROOT . DS . 'plugins' . DS . 'AutoreportModule' . DS . 'webroot' . DS . 'tmp' . DS;
        if (!is_dir($savePath)) {
            return false;
        }

        $files = [];
        foreach (new \DirectoryIterator($savePath) as $fileInfo) {
            if (!$fileInfo->isDot()) {
                $files[] = $fileInfo->getRealPath();
            }
        }
        if (empty($files)) {
            return true;
        }

        $Filesystem = new \Symfony\Component\Filesystem\Filesystem();
        $Filesystem->remove($files);
    }

    /**
     * @param array $timeRanges
     * @return string
     */
    public function weekDaysShort(array $timeRanges) {
        $timeRangeArr = [];
        $weekdaysShort = [
            1 => __('Mo'), // must be __('Monday_short')
            2 => __('Tu'),
            3 => __('We'),
            4 => __('Th'),
            5 => __('Fr'),
            6 => __('Sa'),
            7 => __('Su'),
        ];
        foreach ($timeRanges as $timeRange) {
            $timeRangeArr[] = $weekdaysShort[$timeRange['day']] . ' ' . $timeRange['start'] . ' - ' . $timeRange['end'];
        }
        $timeRangesAsString = implode(' ', $timeRangeArr);
        return $timeRangesAsString;
    }
}
