<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

namespace AutoreportModule\Lib;


use App\Model\Table\SystemfailuresTable;
use AutoreportModule\Model\Table\AutoreportSettingsTable;
use AutoreportModule\Model\Table\AutoreportsTable;
use Cake\Core\Plugin;
use Cake\Http\Exception\NotFoundException;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use itnovum\openITCOCKPIT\Core\DbBackend;
use itnovum\openITCOCKPIT\Core\DowntimeHostConditions;
use itnovum\openITCOCKPIT\Core\DowntimeServiceConditions;
use itnovum\openITCOCKPIT\Core\HoststatusFields;
use itnovum\openITCOCKPIT\Core\NodeJS\ChartRenderClient;
use itnovum\openITCOCKPIT\Core\PerfdataBackend;
use itnovum\openITCOCKPIT\Core\Reports\DaterangesCreator;
use itnovum\openITCOCKPIT\Core\Reports\StatehistoryConverter;
use itnovum\openITCOCKPIT\Core\ServicestatusFields;
use itnovum\openITCOCKPIT\Core\StatehistoryHostConditions;
use itnovum\openITCOCKPIT\Core\StatehistoryServiceConditions;
use itnovum\openITCOCKPIT\Core\Views\Downtime;
use itnovum\openITCOCKPIT\Core\Views\Service;
use itnovum\openITCOCKPIT\Core\Views\UserTime;
use itnovum\openITCOCKPIT\Perfdata\PerfdataLoader;

class Autoreport {


    /**
     * @var array
     */
    private $autoreport;

    /**
     * @var int
     */
    private $startTimestamp;

    /**
     * @var int
     */
    private $endTimestamp;

    /**
     * @var UserTime
     */
    private $UserTime;

    public function __construct(int $id, int $startTimestamp, int $endTimestamp, UserTime $UserTime) {
        /** @var AutoreportsTable $AutoreportsTable */
        $AutoreportsTable = TableRegistry::getTableLocator()->get('AutoreportModule.Autoreports');

        if (!$AutoreportsTable->existsById($id)) {
            throw new NotFoundException();
        }
        $autoreportConfig = $AutoreportsTable->get($id);
        if ($autoreportConfig->get('use_start_time') === 1 && !is_null($autoreportConfig->get('report_start_date'))) {
            $tmpStartTimestamp = $autoreportConfig->get('report_start_date')->getTimestamp();
            if ($tmpStartTimestamp < $endTimestamp) {
                $startTimestamp = strtotime(date('d.m.Y', $tmpStartTimestamp) . ' 00:00:00');
            }

        }

        $this->startTimestamp = ($autoreportConfig->get('use_start_time') === 1 && strtotime($autoreportConfig->get('report_start_date')) < $endTimestamp) ? strtotime($autoreportConfig->get('report_start_date') . ' 00:00:00') : $startTimestamp;
        if ($endTimestamp > time()) {
            $endTimestamp = time();
        }
        $this->endTimestamp = $endTimestamp;
        $this->UserTime = $UserTime;

        $this->autoreport = $AutoreportsTable->getAutoreportByIdForReporting($id, $this->startTimestamp, $this->endTimestamp);

        $weekdays = [
            1 => __('Monday'),
            2 => __('Tuesday'),
            3 => __('Wednesday'),
            4 => __('Thursday'),
            5 => __('Friday'),
            6 => __('Saturday'),
            7 => __('Sunday'),
        ];
        foreach ($this->autoreport['timeperiod']['timeperiod_timeranges'] as $index => $timerange) {
            $this->autoreport['timeperiod']['timeperiod_timeranges'][$index]['weekday'] = $weekdays[$timerange['day']];
        }
    }

    public function canReportBeCreated() {
        if (empty($this->autoreport['timeperiod']['timeperiod_timeranges'])) {
            return [
                'error'   => true,
                'message' => __(
                    'Report "{0}" cannot be evaluated. Time evaluation report data is not available for the selected period "{1}"',
                    $this->autoreport['name'],
                    $this->autoreport['timeperiod']['name']
                ),
                'objects' => []
            ];
        }

        if (empty($this->autoreport['hosts']) && empty($this->autoreport['services'])) {
            return [
                'error'   => true,
                'message' => __(
                    'Report "{0}" cannot be create. No hosts and services are available in selected report',
                    $this->autoreport['name']
                ),
                'objects' => []
            ];
        }

        $timeSlicesGlobal = Hash::insert(
            DaterangesCreator::createDateRanges(
                $this->startTimestamp,
                $this->endTimestamp,
                $this->autoreport['timeperiod']['timeperiod_timeranges']
            ),
            '{n}.is_downtime', false
        );

        $totalTime = Hash::apply(
            array_map(function ($timeSlice) {
                return $timeSlice['end'] - $timeSlice['start'];
            }, $timeSlicesGlobal),
            '{n}',
            'array_sum'
        );

        if ($totalTime === 0) {
            return [
                'error'   => true,
                'message' => __(
                    'The report "{0}" cannot be evaluated. The evaluation time is 0 seconds for the selected period "{1}"',
                    $this->autoreport['name'],
                    $this->autoreport['timeperiod']['name']
                ),
                'objects' => []
            ];
        }

        /** @var AutoreportSettingsTable $AutoreportSettingsTable */
        $AutoreportSettingsTable = TableRegistry::getTableLocator()->get('AutoreportModule.AutoreportSettings');

        $settings = $AutoreportSettingsTable->currentAutoreportSettings();
        $defaultSettings = $AutoreportSettingsTable->defaultAutoreportSettings();

        $checkForNoneExisting = (boolean)($settings['configuration_option'] & $defaultSettings['AUTOREPORTS_GENERATE_ONLY_FOR_EXISTING']['value']);
        if ($checkForNoneExisting) {
            $createdLater = [];
            foreach ($this->autoreport['services'] as $service) {
                //Check if the services was created BEFORE report start date
                if ($this->startTimestamp - $service['created']->getTimestamp() > 0) {
                    continue;
                }

                $hostId = $service['host_id'];
                if (!isset($createdLater[$hostId])) {
                    $createdLater[$hostId]['services'] = [];
                    if ($this->startTimestamp - $service['host']['created']->getTimestamp() > 0) {
                        $createdLater[$hostId]['createdAfter'] = false;
                        $createdLater[$hostId]['name'] = $service['host']['name'];
                    } else {
                        $createdLater[$hostId]['createdAfter'] = true;
                        $createdLater[$hostId]['name'] = $service['host']['name'];
                    }
                }
                $servicename = $service['name'];
                if ($servicename === null || $servicename === '') {
                    $servicename = $service['servicetemplate']['name'];
                }
                $createdLater[$hostId]['services'][] = $servicename;
            }
        }

        if (!empty($createdLater)) {
            return [
                'error'   => true,
                'message' => __(
                    'Report "{0}" cannot be evaluated. The following Hosts/Services were created after the selected starting period "{1}":',
                    $this->autoreport['name'],
                    $this->UserTime->format($this->startTimestamp)
                ),
                'objects' => $createdLater
            ];
        }

        try {
            $timeSlicesGlobal = $this->getGlobalTimeSlices();
        } catch (\Exception $e) {
            return [
                'error'   => true,
                'message' => $e->getMessage(),
                'objects' => []
            ];
        }

        return true;
    }

    /**
     * @return array
     * @throws \Exception
     */
    private function getGlobalTimeSlices() {
        $timeSlicesGlobal = Hash::insert(
            DaterangesCreator::createDateRanges(
                $this->startTimestamp,
                $this->endTimestamp,
                $this->autoreport['timeperiod']['timeperiod_timeranges']
            ),
            '{n}.is_downtime', false
        );

        $timeSlicesCounter = sizeof($timeSlicesGlobal);

        if (!empty($timeSlicesGlobal)) {
            if ($timeSlicesGlobal[0]['start'] < $this->startTimestamp) {
                $timeSlicesGlobal[0]['start'] = $this->startTimestamp;
            }
            if ($timeSlicesGlobal[$timeSlicesCounter - 1]['end'] > $this->endTimestamp) {
                $timeSlicesGlobal[$timeSlicesCounter - 1]['end'] = $this->endTimestamp;
            }
        }


        if ($this->autoreport['calendar_id'] === 0) {
            return $timeSlicesGlobal;
        }

        //filter holidays if needed
        if ($this->autoreport['consider_holidays'] && $this->autoreport['calendar_id'] > 0 && !empty($this->autoreport['calendar']['calendar_holidays'])) {
            $holidays = Hash::extract($this->autoreport['calendar']['calendar_holidays'], '{n}.date');
            foreach ($timeSlicesGlobal as $key => $timeSliceGlobal) {
                if (in_array(date('Y-m-d', $timeSliceGlobal['start']), $holidays, true)) {
                    unset($timeSlicesGlobal[$key]);
                }
            }
            if (empty($timeSlicesGlobal)) {
                throw new \Exception(__(
                    'Time evaluation report data is not available for the selected period ({0} with calendar: {1})',
                    $this->autoreport['timeperiod']['name'],
                    $this->autoreport['calendar']['name']
                ));
            }
        }

        return $timeSlicesGlobal;
    }

    /**
     * @param bool $includeChartImages
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     * @throws \App\Lib\Exceptions\MissingDbBackendException
     */
    public function generateReport($includeChartImages = true) {
        $DbBackend = new DbBackend();
        $PerfdataBackend = new PerfdataBackend();

        $report = $this->calculateReport($DbBackend);

        $GraphImageBlobs = [];
        $configOptionGraph = 1 << 2;

        //Render javascript graphs through NodeJS for PDF and html files
        $ServicestatusFields = new ServicestatusFields($DbBackend);
        $ServicestatusFields->perfdata();

        foreach ($this->autoreport['services'] as $service) {
            if ($service['_joinData']['configuration_option'] & $configOptionGraph) {
                $PerfdataLoader = new PerfdataLoader($DbBackend, $PerfdataBackend);
                switch ($service['_joinData']['graph_settings']) {
                    case 'MIN':
                    case 'min':
                        $seriesAlgorithm = 'min';
                        break;

                    case 'MAX':
                    case 'max':
                        $seriesAlgorithm = 'max';
                        break;

                    default:
                        $seriesAlgorithm = 'avg';
                        break;
                }

                if ($includeChartImages) {
                    // Prometheus is not good as long term storage engine...
                    // openITCOCKPIT has VictoriaMetrics which stores for 1 month by default - also not very optimal
                    if (Plugin::isLoaded('PrometheusModule') && $service['service_type'] === PROMETHEUS_SERVICE) {
                        //Prometheus Services has always a graph
                        $PrometheusPerfdataLoader = new \PrometheusModule\Lib\PrometheusPerfdataLoader();

                        $ServiceObject = new Service($service);

                        $graphData = $PrometheusPerfdataLoader->getPerfdataByUuid(
                            $ServiceObject,
                            $this->startTimestamp,
                            $this->endTimestamp,
                            false
                        );
                    } else {
                        // Whipser graph
                        $graphData = $PerfdataLoader->getPerfdataByUuid(
                            $service['host']['uuid'],
                            $service['uuid'],
                            $this->startTimestamp,
                            $this->endTimestamp,
                            false,
                            $seriesAlgorithm
                        );
                    }

                    if (!empty($graphData)) {
                        $graphServiceName = $service['name'];
                        if (empty($graphServiceName)) {
                            $graphServiceName = $service['servicetemplate']['name'];
                        }

                        //Render graph data to png image blobs for pdf
                        $PuppeteerChartRenderClient = new ChartRenderClient();
                        $PuppeteerChartRenderClient->setGraphStartTimestamp($this->startTimestamp);
                        $PuppeteerChartRenderClient->setGraphEndTimestamp($this->endTimestamp);
                        $PuppeteerChartRenderClient->setTitle(
                            sprintf(
                                '%s - %s',
                                $service['host']['name'],
                                $graphServiceName
                            ));

                        //render two gauges per chart
                        foreach (array_chunk($graphData, 2) as $graphDataChunk) {
                            $GraphImageBlobs[$service['host']['uuid']][$service['uuid']][] = base64_encode($PuppeteerChartRenderClient->getAreaChartAsPngStream($graphDataChunk));
                        }
                    }
                }
            }
        }


        return [
            'autoreport'      => $this->autoreport,
            'autoreport_data' => $report,
            'GraphImageBlobs' => $GraphImageBlobs
        ];
    }

    /**
     * @param DbBackend $DbBackend
     * @return array
     * @throws \App\Lib\Exceptions\MissingDbBackendException
     * @throws \Exception
     */
    private function calculateReport(DbBackend $DbBackend) {
        $HoststatusTable = $DbBackend->getHoststatusTable();
        $ServicestatusTable = $DbBackend->getServicestatusTable();
        $StatehistoryHostsTable = $DbBackend->getStatehistoryHostsTable();
        $StatehistoryServicesTable = $DbBackend->getStatehistoryServicesTable();
        $DowntimehistoryHostsTable = $DbBackend->getDowntimehistoryHostsTable();
        $DowntimehistoryServicesTable = $DbBackend->getDowntimehistoryServicesTable();

        $report = [];
        $timeSlicesGlobal = $this->getGlobalTimeSlices();
        $globalDowntimes = [];

        if ($this->autoreport['consider_downtimes']) {
            /** @var SystemfailuresTable $SystemfailuresTable */
            $SystemfailuresTable = TableRegistry::getTableLocator()->get('Systemfailures');

            $globalDowntimes = $SystemfailuresTable->getSystemfailuresForReportingFormatedAsDowntime(
                $this->startTimestamp,
                $this->endTimestamp
            );
        }
        $totalTime = Hash::apply(Hash::map($timeSlicesGlobal, '{n}', [$this, 'calculateTotalTime']), '{n}', 'array_sum');

        $this->autoreport['total_time'] = $totalTime;

        foreach ($this->autoreport['hosts'] as $host) {
            $hostUuid = $host['uuid'];
            //Process conditions
            $Conditions = new StatehistoryHostConditions();
            $Conditions->setOrder(['StatehistoryHosts.state_time' => 'asc']);
            if ($this->autoreport['check_hard_state']) {
                $Conditions->setHardStateTypeAndUpState(true);
            }

            $report['hosts'][$host['id']]['Host'] = [
                'name'                 => $host['name'],
                'uuid'                 => $host['uuid'],
                'description'          => $host['description'],
                'id'                   => $host['id'],
                'outage_duration'      => $host['_joinData']['outage_duration'],
                'configuration_option' => $host['_joinData']['configuration_option']
            ];

            $Conditions->setFrom($this->startTimestamp);
            $Conditions->setTo($this->endTimestamp);
            $Conditions->setHostUuid($hostUuid);
            $Conditions->setUseLimit(false);

            //Query state history records for hosts
            $statehistories = $StatehistoryHostsTable->getStatehistoryIndex($Conditions, null, false);
            $stateHistoryWithObject[$hostUuid] = [];
            $record = [];
            foreach ($statehistories as $statehistory) {
                $StatehistoryHost = new \itnovum\openITCOCKPIT\Core\Views\StatehistoryHost($statehistory);
                $stateHistoryWithObject[$hostUuid]['Statehistory'][] = $StatehistoryHost->toArray();
            }
            if (empty($stateHistoryWithObject[$hostUuid]['Statehistory'])) {
                //Host has no state history record for selected time range
                //Get last available state history record for this host
                $record = $StatehistoryHostsTable->getLastRecord($Conditions, false);
                if (!empty($record)) {
                    $record['state_time'] = $this->startTimestamp;
                    $StatehistoryHost = new \itnovum\openITCOCKPIT\Core\Views\StatehistoryHost($record);
                    $stateHistoryWithObject[$hostUuid]['Statehistory'][] = $StatehistoryHost->toArray();
                }
            }

            if (empty($stateHistoryWithObject[$hostUuid]['Statehistory']) && empty($record)) {
                $HoststatusFields = new HoststatusFields($DbBackend);
                $HoststatusFields
                    ->currentState()
                    ->isHardstate()
                    ->lastStateChange()
                    ->lastHardState()
                    ->lastHardStateChange();
                $hoststatus = $HoststatusTable->byUuid($hostUuid, $HoststatusFields);

                if (!empty($hoststatus['Hoststatus'])) {
                    $Hoststatus = new \itnovum\openITCOCKPIT\Core\Hoststatus($hoststatus['Hoststatus']);
                    if ($Hoststatus->getLastStateChange() <= $this->startTimestamp) {
                        $record['state_time'] = $this->startTimestamp;
                        $record['state'] = $hoststatus['Hoststatus']['current_state'];
                        $record['last_state'] = $hoststatus['Hoststatus']['current_state'];
                        $record['last_hard_state'] = $hoststatus['Hoststatus']['last_hard_state'];
                        $record['state_type'] = (int)$hoststatus['Hoststatus']['is_hardstate'];
                        $StatehistoryHost = new \itnovum\openITCOCKPIT\Core\Views\StatehistoryHost($record);
                        $stateHistoryWithObject[$hostUuid]['Statehistory'][] = $StatehistoryHost->toArray();
                    }
                }
            }
            if (!empty($stateHistoryWithObject[$hostUuid])) {
                $timeSlices = $timeSlicesGlobal;
                if ($this->autoreport['consider_downtimes']) {
                    //Query downtime records for hosts
                    $DowntimeHostConditions = new DowntimeHostConditions();
                    $DowntimeHostConditions->setOrder(['DowntimeHosts.scheduled_start_time' => 'asc']);
                    $DowntimeHostConditions->setFrom($this->startTimestamp);
                    $DowntimeHostConditions->setTo($this->endTimestamp);
                    $DowntimeHostConditions->setHostUuid($hostUuid);

                    $downtimes = $DowntimehistoryHostsTable->getDowntimesForReporting($DowntimeHostConditions, false, true);
                    foreach ($downtimes as $downtime) {
                        //Use the same format for downtimes and systemdowntimes (also with Statusengine 2 and Statusengine 3)
                        $Downtime = new Downtime($downtime);
                        $endTime = $Downtime->getScheduledEndTime();
                        if ($Downtime->wasCancelled()) {
                            $endTime = $Downtime->getActualEndTime();
                        }

                        $stateHistoryWithObject[$hostUuid]['Downtime'][] = [
                            'author_name'          => $Downtime->getAuthorName(),
                            'comment_data'         => $Downtime->getCommentData(),
                            'entry_time'           => date('Y-m-d H:i:s', $Downtime->getEntryTime()),
                            'scheduled_start_time' => date('Y-m-d H:i:s', $Downtime->getScheduledStartTime()),
                            'scheduled_end_time'   => date('Y-m-d H:i:s', $endTime),
                            'duration'             => $Downtime->getDuration(),
                            'was_started'          => (int)$Downtime->wasStarted(),
                            'internal_downtime_id' => $Downtime->getInternalDowntimeId(),
                            'was_cancelled'        => (int)$Downtime->wasCancelled(),
                        ];
                    }

                    if (!empty($stateHistoryWithObject[$hostUuid]['Downtime'])) {
                        $downtimes = $stateHistoryWithObject[$hostUuid]['Downtime'];
                    }

                    $downtimes = Hash::sort(
                        Hash::filter(
                            array_merge(
                                $globalDowntimes['Systemfailure'],
                                $downtimes
                            )
                        ), '{n}.scheduled_start_time', 'ASC'
                    );

                    if (!empty($downtimes)) {
                        $downtimesFiltered = DaterangesCreator::mergeTimeOverlapping(
                            array_map(
                                function ($downtimes) {
                                    return [
                                        'start_time' => strtotime($downtimes['scheduled_start_time']),
                                        'end_time'   => strtotime($downtimes['scheduled_end_time']),
                                    ];
                                },
                                $downtimes
                            )
                        );

                        $timeSlices = DaterangesCreator::setDowntimesInTimeslices(
                            $timeSlicesGlobal,
                            $downtimesFiltered
                        );
                        unset($downtimesFiltered);
                    }
                }

                $report['hosts'][$host['id']]['Host'] = array_merge(
                    $report['hosts'][$host['id']]['Host'],
                    StatehistoryConverter::generateReportDataWithOutages(
                        $timeSlices,
                        $stateHistoryWithObject[$hostUuid]['Statehistory'],
                        $this->autoreport['check_hard_state'],
                        true
                    )
                );

                $states = [
                    0 => 'up',
                    1 => 'down',
                    2 => 'unreachable'
                ];
                if ($totalTime > 0) {
                    foreach ($states as $stateId => $stateName) {
                        $statePercentage = $report['hosts'][$host['id']]['Host'][$stateId] / $totalTime * 100;
                        $report['hosts'][$host['id']]['Host']['state_overview']['percent_' . $stateName] = $statePercentage;
                        $report['hosts'][$host['id']]['Host']['state_overview']['human_' . $stateName] = $this->UserTime->secondsInHumanShort($report['hosts'][$host['id']]['Host'][$stateId]);
                    }
                    $report['hosts'][$host['id']]['Host']['state_overview']['percent_availability'] = 100 - $report['hosts'][$host['id']]['Host']['state_overview']['percent_down'];

                    $report['hosts'][$host['id']]['Host']['downtimes'] = $downtimes ?? [];

                    $report['hosts'][$host['id']]['Host']['frontend'] = $this->getOutagesForFrontend($report['hosts'][$host['id']]['Host']);
                }

            }
        }

        foreach ($this->autoreport['services'] as $service) {
            if (!isset($report['hosts'][$service['host_id']]['Host'])) {
                $report['hosts'][$service['host_id']]['Host'] = $service['host'];
            }
            $report['hosts'][$service['host_id']]['Services'][$service['id']]['Service'] = [
                'name'                 => ($service['name']) ? $service['name'] : $service['servicetemplate']['name'],
                'uuid'                 => $service['uuid'],
                'description'          => ($service['description']) ? $service['description'] : $service['servicetemplate']['description'],
                'id'                   => $service['id'],
                'outage_duration'      => $service['_joinData']['outage_duration'],
                'configuration_option' => $service['_joinData']['configuration_option'],
                'graph_settings'       => $service['_joinData']['graph_settings']
            ];
            $record = [];
            $serviceUuid = $service['uuid'];
            //Process conditions
            $Conditions = new StatehistoryServiceConditions();
            $Conditions->setOrder(['StatehistoryServices.state_time' => 'asc']);
            if ($this->autoreport['check_hard_state']) {
                $Conditions->setHardStateTypeAndOkState(true);
            }
            $Conditions->setFrom($this->startTimestamp);
            $Conditions->setTo($this->endTimestamp);
            $Conditions->setServiceUuid($serviceUuid);
            $Conditions->setUseLimit(false);

            //Query state history records for services
            $statehistories = $StatehistoryServicesTable->getStatehistoryIndex($Conditions, null, false);
            $stateHistoryWithObject[$serviceUuid] = [];
            foreach ($statehistories as $statehistory) {
                $StatehistoryService = new \itnovum\openITCOCKPIT\Core\Views\StatehistoryService($statehistory);
                $stateHistoryWithObject[$serviceUuid]['Statehistory'][] = $StatehistoryService->toArray();
            }
            if (empty($stateHistoryWithObject[$serviceUuid]['Statehistory'])) {
                //Service has no state history record for selected time range
                //Get last available state history record for this service
                $record = $StatehistoryServicesTable->getLastRecord($Conditions, false);
                if (!empty($record)) {
                    $record['state_time'] = $this->startTimestamp;
                    $StatehistoryService = new \itnovum\openITCOCKPIT\Core\Views\StatehistoryService($record);
                    $stateHistoryWithObject[$serviceUuid]['Statehistory'][] = $StatehistoryService->toArray();
                }
            }

            if (empty($stateHistoryWithObject[$serviceUuid]['Statehistory']) && empty($record)) {
                $ServicestatusFields = new ServicestatusFields($DbBackend);
                $ServicestatusFields
                    ->currentState()
                    ->isHardstate()
                    ->lastHardState()
                    ->lastStateChange()
                    ->lastHardStateChange();
                $servicestatus = $ServicestatusTable->byUuid($service['uuid'], $ServicestatusFields);
                if (!empty($servicestatus['Servicestatus'])) {
                    $Servicestatus = new \itnovum\openITCOCKPIT\Core\Servicestatus($servicestatus['Servicestatus']);
                    if ($Servicestatus->getLastStateChange() <= $this->startTimestamp) {
                        $record['state_time'] = $this->startTimestamp;
                        $record['state'] = $servicestatus['Servicestatus']['current_state'];
                        $record['last_state'] = $servicestatus['Servicestatus']['current_state'];
                        $record['last_hard_state'] = $servicestatus['Servicestatus']['last_hard_state'];
                        $record['state_type'] = (int)$servicestatus['Servicestatus']['is_hardstate'];
                        $StatehistoryService = new \itnovum\openITCOCKPIT\Core\Views\StatehistoryService($record);
                        $stateHistoryWithObject[$serviceUuid]['Statehistory'][] = $StatehistoryService->toArray();
                    }
                }
            }


            if (!empty($stateHistoryWithObject[$serviceUuid])) {
                $timeSlices = $timeSlicesGlobal;
                if ($this->autoreport['consider_downtimes']) {
                    //Query downtime records for services
                    $DowntimeServiceConditions = new DowntimeServiceConditions();
                    $DowntimeServiceConditions->setOrder(['DowntimeServices.scheduled_start_time' => 'asc']);
                    $DowntimeServiceConditions->setFrom($this->startTimestamp);
                    $DowntimeServiceConditions->setTo($this->endTimestamp);
                    $DowntimeServiceConditions->setServiceUuid($serviceUuid);

                    $downtimes = $DowntimehistoryServicesTable->getDowntimesForReporting($DowntimeServiceConditions, false, true);

                    foreach ($downtimes as $downtime) {
                        //Make system downtimes and downtimes looks like the same also with Statusengine 2 and 3
                        $Downtime = new Downtime($downtime);
                        $endTime = $Downtime->getScheduledEndTime();
                        if ($Downtime->wasCancelled()) {
                            $endTime = $Downtime->getActualEndTime();
                        }

                        $stateHistoryWithObject[$serviceUuid]['Downtime'][] = [
                            'author_name'          => $Downtime->getAuthorName(),
                            'comment_data'         => $Downtime->getCommentData(),
                            'entry_time'           => date('Y-m-d H:i:s', $Downtime->getEntryTime()),
                            'scheduled_start_time' => date('Y-m-d H:i:s', $Downtime->getScheduledStartTime()),
                            //'scheduled_end_time' => date('Y-m-d H:i:s', $Downtime->getScheduledEndTime()),
                            'scheduled_end_time'   => date('Y-m-d H:i:s', $endTime),
                            'duration'             => $Downtime->getDuration(),
                            'was_started'          => (int)$Downtime->wasStarted(),
                            'internal_downtime_id' => $Downtime->getInternalDowntimeId(),
                            'was_cancelled'        => (int)$Downtime->wasCancelled(),
                        ];
                    }

                    if (!empty($stateHistoryWithObject[$serviceUuid]['Downtime'])) {
                        $downtimes = $stateHistoryWithObject[$serviceUuid]['Downtime'];
                    }

                    $downtimes = Hash::sort(
                        Hash::filter(
                            array_merge(
                                $globalDowntimes['Systemfailure'],
                                $downtimes
                            )
                        ), '{n}.scheduled_start_time', 'ASC'
                    );

                    if (!empty($downtimes)) {
                        $downtimesFiltered = DaterangesCreator::mergeTimeOverlapping(
                            array_map(
                                function ($downtimes) {
                                    return [
                                        'start_time' => strtotime($downtimes['scheduled_start_time']),
                                        'end_time'   => strtotime($downtimes['scheduled_end_time']),
                                    ];
                                },
                                $downtimes
                            )
                        );
                        $timeSlices = DaterangesCreator::setDowntimesInTimeslices(
                            $timeSlicesGlobal,
                            $downtimesFiltered
                        );
                        unset($downtimesFiltered);
                    }
                }

                $report['hosts'][$service['host_id']]['Services'][$service['id']]['Service'] = array_merge(
                    $report['hosts'][$service['host_id']]['Services'][$service['id']]['Service'],
                    StatehistoryConverter::generateReportDataWithOutages(
                        $timeSlices,
                        $stateHistoryWithObject[$serviceUuid]['Statehistory'],
                        $this->autoreport['check_hard_state'],
                        false
                    )
                );

                $states = [
                    0 => 'ok',
                    1 => 'warning',
                    2 => 'critical',
                    3 => 'unknown'
                ];
                if ($totalTime > 0) {
                    foreach ($states as $stateId => $stateName) {
                        $statePercentage = $report['hosts'][$service['host_id']]['Services'][$service['id']]['Service'][$stateId] / $totalTime * 100;
                        $report['hosts'][$service['host_id']]['Services'][$service['id']]['Service']['state_overview']['percent_' . $stateName] = $statePercentage;
                        $report['hosts'][$service['host_id']]['Services'][$service['id']]['Service']['state_overview']['human_' . $stateName] = $this->UserTime->secondsInHumanShort($report['hosts'][$service['host_id']]['Services'][$service['id']]['Service'][$stateId]);
                    }
                    $report['hosts'][$service['host_id']]['Services'][$service['id']]['Service']['state_overview']['percent_availability'] = 100 - $report['hosts'][$service['host_id']]['Services'][$service['id']]['Service']['state_overview']['percent_critical'];

                    $report['hosts'][$service['host_id']]['Services'][$service['id']]['Service']['downtimes'] = $downtimes ?? [];

                    $report['hosts'][$service['host_id']]['Services'][$service['id']]['Service']['frontend'] = $this->getOutagesForFrontend($report['hosts'][$service['host_id']]['Services'][$service['id']]['Service']);
                }
            }

        }
        /** @var AutoreportSettingsTable $AutoreportSettingsTable */
        $AutoreportSettingsTable = TableRegistry::getTableLocator()->get('AutoreportModule.AutoreportSettings');

        $settings = $AutoreportSettingsTable->currentAutoreportSettings();
        $defaultSettings = $AutoreportSettingsTable->defaultAutoreportSettings();

        if (!empty($report['hosts'])) {
            $outages_filtered = StatehistoryConverter::mergeOutages(
                Hash::sort(
                    Hash::merge(
                        Hash::extract($report['hosts'], '{n}.Host.outages.{n}'),
                        Hash::extract($report['hosts'], '{n}.Services.{n}.Service.outages.{n}')
                    ),
                    '{n}.start', 'ASC'
                )
            );


            $total_outage_time = Hash::apply(Hash::map($outages_filtered, '{n}', [$this, 'calculateTotalTime']), '{n}', 'array_sum');

            $report['total_time'] = $totalTime;
            $report['total_outage_time'] = $total_outage_time;
            $report['outages_filtered'] = $outages_filtered;
            $report['time_slices'] = $timeSlicesGlobal;
            $report['last_number_of_outages'] = $this->autoreport['last_number_of_outages'];
            $report['evaluation_time'] = [
                'start'       => $this->startTimestamp,
                'start_human' => $this->UserTime->format($this->startTimestamp),
                'end'         => $this->endTimestamp,
                'end_human'   => $this->UserTime->format($this->endTimestamp)
            ];

            if ($totalTime > 0) {
                if ($this->autoreport['min_availability_percent'] === true) { // Guaranteed availability
                    $report['max_outages_time_in_percent'] = 100 - $this->autoreport['min_availability'];
                    $report['max_outages_time_in_seconds'] = ($report['max_outages_time_in_percent'] / 100 * $totalTime);
                    $report['max_outages_time_in_seconds_human'] = $this->UserTime->secondsInHumanShort((int)$report['max_outages_time_in_seconds']);
                } else {
                    // min_availability is stored in minutes so * 60 to get seconds
                    $report['max_outages_time_in_percent'] = ($this->autoreport['min_availability'] * 60) / $totalTime * 100;
                    $report['max_outages_time_in_seconds'] = $this->autoreport['min_availability'] * 60;
                    $report['max_outages_time_in_seconds_human'] = $this->UserTime->secondsInHumanShort((int)$report['max_outages_time_in_seconds']);
                }

                $report['total_outage_time_percentage'] = $report['total_outage_time'] / $totalTime * 100;
                $report['total_outage_time_human'] = $this->UserTime->secondsInHumanShort((int)$report['total_outage_time']);
                $report['total_outage_count'] = sizeof($outages_filtered);


                $report['sla_graph_path'] = null;
                if (($settings['configuration_option'] & $defaultSettings['AUTOREPORTS_SHOW_SLA_GRAPH']['value'])) {
                    $SlaChart = new SlaChart($this->UserTime);
                    $report['sla_graph_path'] = $SlaChart->getReportDataAsPng(
                        $this->startTimestamp,
                        $this->endTimestamp,
                        $this->autoreport['report_interval'],
                        $report['time_slices'],
                        $this->autoreport['timeperiod']['timeperiod_timeranges'],
                        $report['outages_filtered'],
                        $report['max_outages_time_in_percent']
                    );
                }
            }
        }
        $report['settings']['configuration_option'] = (int)$settings['configuration_option'];
        $report['defaultSettings'] = $defaultSettings;

        return $report;
    }

    /**
     * @param array $time_slice
     * @return int
     */
    public function calculateTotalTime($time_slice) {
        return $time_slice['end'] - $time_slice['start'];
    }


    /**
     * @param array $objectWithOutages
     * @return array
     * @throws \Exception
     */
    public function getOutagesForFrontend($objectWithOutages) {
        $formated = [
            'outages'                        => [],
            'outages_summarized_in_downtime' => [],
            'downtimes'                      => [],
            'maximum_outage_duration'        => intval($objectWithOutages['maximum_outage_duration'] ?? 0),
        ];

        $formated['show_outages'] = false;
        if ($objectWithOutages['configuration_option'] & 8) { //8 === show ALL outage
            $formated['outages_headline'] = __('Outages');
            $formated['show_outages'] = true;
        } else {
            if ($objectWithOutages['outage_duration'] !== null) {
                $formated['outages_headline'] = __('Outages of at least {0} min', $objectWithOutages['outage_duration']);
                $formated['show_outages'] = true;
            }
        }

        foreach ($objectWithOutages['outages'] as $outage) {
            $report_outage_duration_in_seconds = $objectWithOutages['outage_duration'];
            if ($report_outage_duration_in_seconds !== null) {
                $report_outage_duration_in_seconds = $report_outage_duration_in_seconds * 60;
            }

            $duration = $outage['end'] - $outage['start'];
            if ($duration >= $report_outage_duration_in_seconds || ($objectWithOutages['configuration_option'] & 8)) { //8 === show ALL outtages
                if ($duration > $report_outage_duration_in_seconds && $report_outage_duration_in_seconds !== null) {
                    $durationStr = sprintf(
                        '%s (+ %s)',
                        $this->UserTime->secondsInHumanShort($report_outage_duration_in_seconds),
                        $this->UserTime->secondsInHumanShort($duration - $report_outage_duration_in_seconds)
                    );
                } else {
                    $durationStr = sprintf(
                        '%s',
                        $this->UserTime->secondsInHumanShort($duration)
                    );
                }

                $formated['outages'][] = [
                    'start'        => $this->UserTime->format($outage['start']),
                    'end'          => $this->UserTime->format($outage['end']),
                    'duration_str' => $durationStr,
                    'duration'     => (int)$duration,
                ];
            }
        }

        foreach ($objectWithOutages['outages_summarized_in_downtime'] as $outageInDowntime) {
            $formated['outages_summarized_in_downtime'][] = [
                'start'        => $this->UserTime->format($outageInDowntime['start']),
                'end'          => $this->UserTime->format($outageInDowntime['end']),
                'duration_str' => $this->UserTime->secondsInHumanShort($outageInDowntime['end'] - $outageInDowntime['start'])
            ];
        }

        foreach ($objectWithOutages['downtimes'] as $downtime) {
            $start = strtotime($downtime['scheduled_start_time']);
            $end = strtotime($downtime['scheduled_end_time']);

            $formated['downtimes'][] = [
                'start'        => $this->UserTime->format($start),
                'end'          => $this->UserTime->format($end),
                'duration_str' => $this->UserTime->secondsInHumanShort($end - $start),
                'author_name'  => $downtime['author_name'],
                'comment_data' => $downtime['comment_data']
            ];
        }

        $formated['maximum_outage_duration_human'] = $this->UserTime->secondsInHumanShort($formated['maximum_outage_duration']);
        return $formated;
    }

}