<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

namespace AutoreportModule\Model\Table;

use AutoreportModule\Model\Entity\AutoreportAvailabilityLog;
use Cake\Http\Exception\NotFoundException;
use Cake\ORM\Table;
use Cake\ORM\TableRegistry;
use itnovum\openITCOCKPIT\Core\Reports\DaterangesCreator;
use itnovum\openITCOCKPIT\Core\Views\UserTime;

class AutoreportAvailabilityLogTable extends Table {
    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);
        $this->addBehavior('Timestamp');

        $this->setEntityClass('AutoreportModule.AutoreportAvailabilityLog');
        $this->setTable('autoreports_availability_log');

        $this->belongsTo('AutoreportModule.Autoreports');
    }

    /**
     * @param $autoreportId
     * @return array|\Cake\Datasource\EntityInterface|null
     */
    public function getLastAutoreportAvailabilityLogByAutoreportId($autoreportId) {
        $result = $this->find()
            ->where([
                'AutoreportAvailabilityLog.autoreport_id' => $autoreportId
            ])
            ->order(['AutoreportAvailabilityLog.id' => 'desc'])
            ->disableHydration()
            ->limit(1)
            ->first();

        if (empty($result)) {
            return [];
        }
        return $result;
    }

    /**
     * @param $reportId
     * @return array|void
     * @throws \App\Lib\Exceptions\MissingDbBackendException
     * @throws \GuzzleHttp\Exception\GuzzleException
     * @throws \Exception
     */
    public function calculateAvailabilityForReport($reportId) {
        /** @var $AutoreportsTable AutoreportsTable */
        $AutoreportsTable = TableRegistry::getTableLocator()->get('AutoreportModule.Autoreports');
        if (!$AutoreportsTable->existsByIdForAvailabilityCalculation($reportId)) {
            throw new NotFoundException(__('Auto report not found'));
        }

        $autoreport = $AutoreportsTable->getAutoreportById($reportId);
        $calculatedTimeRanges = [];
        if (!empty($autoreport)) {
            $ServerTime = new \DateTime();
            $ServerTimeZone = new \DateTimeZone($ServerTime->getTimezone()->getName());


            $UserTimeWithServerTime = new UserTime($ServerTimeZone->getName(), 'H:i:s - d.m.Y');
            $firstRun = true;
            $initStartDate = DaterangesCreator::getInitStartTimeForCalculateAvailability($autoreport->get('report_interval'));
            $lastReportLogEntry = $this->getLastAutoreportAvailabilityLogByAutoreportId($reportId);

            if (!empty($lastReportLogEntry)) {
                $initStartDate = $lastReportLogEntry['last_update'];
                $firstRun = false;
            }
            $dateRanges = DaterangesCreator::createDateSlicesByIntervalAndLastUpdateDate(
                $initStartDate,
                $autoreport->get('report_interval')
            );

            foreach ($dateRanges as $dateRange) {
                $start = $dateRange['start'];
                $end = $dateRange['end'];
                $entryType = $dateRange['entryType'];
                $lastReportLogEntry = [];
                if (!$firstRun) {
                    $lastReportLogEntry = $this->getLastAutoreportAvailabilityLogByAutoreportId($reportId);
                }
                $Autoreport = new \AutoreportModule\Lib\Autoreport($reportId, $start, $end, $UserTimeWithServerTime);
                $result = $Autoreport->canReportBeCreated();
                if ($result === true) {
                    $report = $Autoreport->generateReport(false);

                    if (!empty($report['autoreport_data']) && $report['autoreport_data']['total_time'] > 0) {
                        $currentTotalTime = $report['autoreport_data']['total_time'];
                        $currentTotalOutageTime = $report['autoreport_data']['total_outage_time'];
                        $currentAvailabilityInPercent = strval(round(100 * ($currentTotalTime - $currentTotalOutageTime) / $currentTotalTime, 3));

                        $availabilityRecord = [
                            'AutoreportAvailabilityLog' => [
                                'autoreport_id'                   => $reportId,
                                'report_interval'                 => $autoreport->get('report_interval'),
                                'last_update'                     => $end,
                                'evaluation_start'                => $start,
                                'evaluation_end'                  => $end,
                                'availability_total_time'         => $currentTotalTime,
                                'minimal_availability_time'       => (int)($currentTotalTime * $report['autoreport']['min_availability'] / 100),
                                'determined_availability_time'    => $currentTotalTime - $currentTotalOutageTime,
                                'minimal_availability_percent'    => $report['autoreport']['min_availability'],
                                'determined_availability_percent' => $currentAvailabilityInPercent,
                                'determined_outage_time'          => $currentTotalOutageTime,
                                'maximal_number_outages'          => $report['autoreport']['max_number_of_outages'],
                                'determined_number_outages'       => sizeof($report['autoreport_data']['outages_filtered'])
                            ]
                        ];

                        if ($entryType === 'update' && !empty($lastReportLogEntry) && !empty($report['autoreport_data'])) {
                            $totalTime = $currentTotalTime + $lastReportLogEntry['availability_total_time'];
                            $totalOutageTime = $currentTotalOutageTime + $lastReportLogEntry['determined_outage_time'];
                            $availabilityInPercent = strval(round(100 * ($totalTime - $totalOutageTime) / $totalTime, 3));

                            $availabilityRecord['AutoreportAvailabilityLog']['availability_total_time'] += $lastReportLogEntry['availability_total_time'];
                            $availabilityRecord['AutoreportAvailabilityLog']['determined_availability_percent'] = $availabilityInPercent;
                            $availabilityRecord['AutoreportAvailabilityLog']['minimal_availability_time'] += $lastReportLogEntry['minimal_availability_time'];
                            $availabilityRecord['AutoreportAvailabilityLog']['determined_availability_time'] += $lastReportLogEntry['determined_availability_time'];
                            $availabilityRecord['AutoreportAvailabilityLog']['determined_outage_time'] = $totalOutageTime;
                            $availabilityRecord['AutoreportAvailabilityLog']['determined_number_outages'] += $lastReportLogEntry['determined_number_outages'];
                        }
                        if (!empty($availabilityRecord)) {
                            /** @var  $autoreportAvailabilityLog AutoreportAvailabilityLog */
                            $autoreportAvailabilityLog = $this->newEmptyEntity();
                            $autoreportAvailabilityLog = $this->patchEntity(
                                $autoreportAvailabilityLog,
                                $availabilityRecord
                            );

                            $this->save($autoreportAvailabilityLog);
                            if ($autoreportAvailabilityLog->hasErrors()) {
                                return [
                                    'hasError' => true,
                                    'data'     => $autoreportAvailabilityLog
                                ];
                            };

                            $calculatedTimeRanges[] = [
                                'start' => $start,
                                'end'   => $end
                            ];
                        }
                    }
                }
            }
        }
        return $calculatedTimeRanges;
    }
}
