<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace AutoreportModule\Model\Table;

use Cake\ORM\Table;
use Cake\Validation\Validator;

/**
 * EventcorrelationSettings Model
 *
 * @method \AutoreportModule\Model\Entity\AutoreportSetting get($primaryKey, $options = [])
 * @method \AutoreportModule\Model\Entity\AutoreportSetting newEntity($data = null, array $options = [])
 * @method \AutoreportModule\Model\Entity\AutoreportSetting[] newEntities(array $data, array $options = [])
 * @method \AutoreportModule\Model\Entity\AutoreportSetting|bool save(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \AutoreportModule\Model\Entity\AutoreportSetting saveOrFail(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \AutoreportModule\Model\Entity\AutoreportSetting patchEntity(\Cake\Datasource\EntityInterface $entity, array $data, array $options = [])
 * @method \AutoreportModule\Model\Entity\AutoreportSetting[] patchEntities($entities, array $data, array $options = [])
 * @method \AutoreportModule\Model\Entity\AutoreportSetting findOrCreate($search, callable $callback = null, $options = [])
 *
 * @mixin \Cake\ORM\Behavior\TimestampBehavior
 */
class AutoreportSettingsTable extends Table {
    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('autoreport_settings');
        $this->setDisplayField('id');
        $this->setPrimaryKey('id');

        $this->addBehavior('Timestamp');
    }

    /**
     * Default validation rules.
     *
     * @param \Cake\Validation\Validator $validator Validator instance.
     * @return \Cake\Validation\Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $validator
            ->integer('id')
            ->allowEmptyString('id', 'create');

        $validator
            ->scalar('store_path')
            ->allowEmptyString('store_path')
            ->regex('store_path', '/^((?:\/[a-zA-Z0-9]+(?:_[a-zA-Z0-9]+)*(?:\-[a-zA-Z0-9]+)*)+)$/', __('Please enter a valid path'));

        $validator
            ->scalar('notification_emails')
            ->add('notification_emails', 'custom', [
                'rule'    => function ($value, $context) {
                    if (empty($value)) {
                        return true;
                    }
                    $emails = explode(';', $value);
                    foreach ($emails as $email) {
                        if (!filter_var($email, FILTER_VALIDATE_EMAIL)) {
                            return false;
                        }
                    }
                    return true;
                },
                'message' => 'Please enter valid email addresses. Use ";" as email delimiter when sending to multiple recipients'
            ]);

        return $validator;
    }


    /**
     * @return array
     */
    public function currentAutoreportSettings() {
        $default = [
            'configuration_option' => 1 << 6 | 1 << 7 | 1 << 8,
            'store_path'           => '',
            'notification_emails'  => ''
        ];

        $result = $this->find()
            ->where([
                'id' => 1
            ])
            ->disableHydration()
            ->first();

        if (empty($result)) {
            return $default;
        }

        return $result;

    }


    /**
     * @return \Cake\Datasource\EntityInterface
     */
    public function getAutoreportSettingsEntity() {
        $result = $this->find()
            ->where([
                'id' => 1
            ])
            ->first();
        if (empty($result)) {
            $entity = $this->newEmptyEntity();
            $entity->set('id', 1);
            $entity->set('configuration_option', 1 << 6 | 1 << 7 | 1 << 8);
            $entity->set('store_path', '');
            $entity->set('notification_emails', '');
            $entity->setAccess('id', false);
            return $entity;
        }

        $result->setAccess('id', false);
        return $result;
    }

    /**
     * @return array
     */
    public function defaultAutoreportSettings() {
        $all_autoreport_settings = [
            'AUTOREPORTS_TENANT_AS_SUBJECT'           => [
                'value' => 1 << 0,
                'info'  => __('Use tenant name instead of report name as subject for automatic generated mails'),
            ],
            'AUTOREPORTS_GENERATE_ONLY_FOR_EXISTING'  => [
                'value' => 1 << 11,
                'info'  => __('Generates reports only for Hosts and Services that were created before the start date'),
            ],
            'AUTOREPORTS_SEND_REPORT_AS_PDF'          => [
                'value' => 1 << 1,
                'info'  => __('Send reports in PDF format'),
            ],
            'AUTOREPORTS_SEND_REPORT_AS_CSV'          => [
                'value' => 1 << 9,
                'info'  => __('Send reports in CSV format'),
            ],
            'AUTOREPORTS_GENERATE_QUARTERLY_REPORTS'  => [
                'value' => 1 << 10,
                'info'  => __('Send quarterly reports'),
            ],
            'AUTOREPORTS_SHOW_OUTAGES_IN_DOWNTIME'    => [
                'value' => 1 << 2,
                'info'  => __('Shows outages in scheduled downtime'),
            ],
            'AUTOREPORTS_SHOW_DOWNTIMES'              => [
                'value' => 1 << 3,
                'info'  => __('Shows all downtimes'),
            ],
            'AUTOREPORTS_USE_CREATE_DATE_IN_FILENAME' => [
                'value' => 1 << 12,
                'info'  => __('Use create date in filename. Format: Filename "Reportname_+/- (dmY)..."'),
            ],
            'AUTOREPORTS_USE_YEAR_IN_FILENAME'        => [
                'value' => 1 << 4,
                'info'  => __('Use Year in filename. Format: Filename "Reportname_+/- (Y)..."'),
            ],
            'AUTOREPORTS_SHOW_SLA_GRAPH'              => [
                'value' => 1 << 5,
                'info'  => __('Displays SLA graph'),
            ],
            'AUTOREPORTS_SHOW_SUMMARY_STATISTICS'     => [
                'value' => 1 << 6,
                'info'  => __('Shows summary statistics on first page of report'),
            ],
            'AUTOREPORTS_EXTENDED_HOST_DESC'          => [
                'value' => 1 << 7,
                'info'  => __('Shows host name with current availability'),
            ],
            'AUTOREPORTS_EXTENDED_SERV_DESC'          => [
                'value' => 1 << 8,
                'info'  => __('Shows service description with current availability'),
            ],
            'AUTOREPORTS_ERROR_NOTIFICATION'          => [
                'value' => '',
                'info'  => __('Email addresses which will be notified in case of failure while creating the quarterly reports. Use ";" as email delimiter when sending to multiple recipients'),
            ],
            'AUTOREPORTS_STORE_PATH'                  => [
                'value' => '',
                'info'  => __('Path for Reports files'),
            ],
        ];

        return $all_autoreport_settings;
    }
}
