<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace AutoreportModule\Model\Table;

use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Model\Table\HostsTable;
use App\Model\Table\ServicesTable;
use AutoreportModule\Filter\AutoreportFilter;
use AutoreportModule\Lib\AutoreportServiceConditions;
use AutoreportModule\Model\Entity\Autoreport;
use Cake\ORM\Query;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use Cake\Validation\Validator;
use itnovum\openITCOCKPIT\Database\PaginateOMat;

/**
 * Autoreports Model
 *
 * @property \App\Model\Table\HostsTable|\Cake\ORM\Association\BelongsTo $Hosts
 * @property \App\Model\Table\ServicesTable|\Cake\ORM\Association\BelongsTo $Services
 *
 * @method \AutoreportModule\Model\Entity\Autoreport get($primaryKey, $options = [])
 * @method \AutoreportModule\Model\Entity\Autoreport newEntity($data = null, array $options = [])
 * @method \AutoreportModule\Model\Entity\Autoreport[] newEntities(array $data, array $options = [])
 * @method \AutoreportModule\Model\Entity\Autoreport|bool save(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \AutoreportModule\Model\Entity\Autoreport saveOrFail(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \AutoreportModule\Model\Entity\Autoreport patchEntity(\Cake\Datasource\EntityInterface $entity, array $data, array $options = [])
 * @method \AutoreportModule\Model\Entity\Autoreport[] patchEntities($entities, array $data, array $options = [])
 * @method \AutoreportModule\Model\Entity\Autoreport findOrCreate($search, callable $callback = null, $options = [])
 *
 * @mixin \Cake\ORM\Behavior\TreeBehavior
 */
class AutoreportsTable extends Table {
    use PaginationAndScrollIndexTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('autoreports');
        $this->setDisplayField('id');
        $this->setPrimaryKey('id');

        $this->addBehavior('Timestamp');

        $this->belongsTo('Containers', [
            'foreignKey' => 'container_id',
            'joinType'   => 'INNER'
        ]);

        $this->belongsTo('Timeperiods', [
            'foreignKey' => 'timeperiod_id',
            'joinType'   => 'INNER'
        ]);

        $this->belongsTo('Calendars', [
            'foreignKey' => 'calendar_id',
            'joinType'   => 'LEFT'
        ]);

        $this->belongsToMany('Users', [
            'joinTable'    => 'users_to_autoreports',
            'saveStrategy' => 'replace'
        ])->setDependent(true);


        $this->belongsToMany('Hosts', [
            'through'          => 'AutoreportModule.AutoreportsHostsMemberships',
            'className'        => 'Hosts',
            'foreignKey'       => 'autoreport_id',
            'targetForeignKey' => 'host_id',
            'joinTable'        => 'hosts_to_autoreports',
            'saveStrategy'     => 'replace'
        ])->setDependent(true);

        $this->belongsToMany('Services', [
            'through'          => 'AutoreportModule.AutoreportsServicesMemberships',
            'className'        => 'Services',
            'foreignKey'       => 'autoreport_id',
            'targetForeignKey' => 'service_id',
            'joinTable'        => 'services_to_autoreports',
            'saveStrategy'     => 'replace'
        ])->setDependent(true);

        $this->hasMany('AutoreportAvailabilityLog', [
            'className' => 'AutoreportModule.AutoreportAvailabilityLog'

        ])->setDependent(true);

    }

    /**
     * Default validation rules.
     *
     * @param \Cake\Validation\Validator $validator Validator instance.
     * @return \Cake\Validation\Validator
     */
    public function validationAddStepOne(Validator $validator): Validator {
        $validator
            ->integer('id')
            ->allowEmptyString('id', null, 'create');

        $validator
            ->integer('container_id')
            ->requirePresence('container_id', 'create')
            ->allowEmptyString('container_id', null, false)
            ->greaterThanOrEqual('container_id', 1);

        $validator
            ->scalar('name')
            ->maxLength('name', 255)
            ->requirePresence('name', 'create')
            ->notEmptyString('name');

        $validator
            ->scalar('timeperiod_id')
            ->requirePresence('timeperiod_id', 'create')
            ->notEmptyString('timeperiod_id');

        $validator
            ->requirePresence('report_interval', 'create')
            ->allowEmptyString('report_interval', null, false);

        $validator
            ->requirePresence('report_send_interval', 'create')
            ->allowEmptyString('report_interval', null, false);

        $validator
            ->date('report_start_date', ['ymd'])
            ->allowEmptyString('report_start_date', __('This field cannot be left blank'),
                function ($context) {
                    if ($context['data']['use_start_time'] === 0) {
                        return true;
                    }
                    return false;
                }
            );

        $validator
            ->regex('min_availability', '/(^100(\.0{1,3})?$)|(^([1-9]([0-9])?|0)(\.[0-9]{1,3})?$)/',
                __('The provided value is invalid (For example 99.999)'),
                function ($context) {
                    if ($context['data']['min_availability_percent'] === true) {
                        return true;
                    }
                    return false;
                }
            )
            ->nonNegativeInteger('min_availability',
                __('The provided value must be a number'),
                function ($context) {
                    if ($context['data']['min_availability_percent'] === true) {
                        return false;
                    }
                    return true;
                }
            )
            ->range('min_availability', [0, 100], __('The provided value is invalid (min: 0 max: 100)'),
                function ($context) {
                    if ($context['data']['min_availability_percent']) {
                        return true;
                    }
                    return false;
                }
            )
            ->allowEmptyString('min_availability', __('This field cannot be left blank'),
                function ($context) {
                    if (!empty($context['data']['max_number_of_outages'])) {
                        return true;
                    }
                    return false;
                }
            );
        $validator
            ->greaterThanOrEqual('max_number_of_outages', 0)
            ->allowEmptyString('max_number_of_outages', __('This field cannot be left blank'),
                function ($context) {
                    if (!empty($context['data']['min_availability'])) {
                        return true;
                    }
                    return false;
                }
            );

        $validator
            ->add('users', 'custom', [
                'rule'    => [$this, 'atLeastOne'],
                'message' => __('You must select at least one user.')
            ]);

        $validator
            ->greaterThan('calendar_id', 0, __('Please select a calendar'))
            ->allowEmptyString('calendar_id', null, function ($context) {
                if ($context['data']['consider_holidays'] == 0) {
                    return true;
                }
                return false;
            });


        return $validator;
    }

    /**
     * Default validation rules.
     *
     * @param \Cake\Validation\Validator $validator Validator instance.
     * @return \Cake\Validation\Validator
     */
    public function validationAddStepTwo(Validator $validator): Validator {
        $validator
            ->integer('id')
            ->allowEmptyString('id', null, 'create');

        $validator
            ->notEmptyArray('hosts')
            ->add('hosts', 'custom', [
                'rule'    => [$this, 'checkOutageDuration'],
                'message' => __('Outage duration needs to be numeric')
            ])
            ->add('hosts', 'custom', [
                'rule'    => [$this, 'atLeastOneHostOrServiceInReport'],
                'message' => __('You must select at least one option for host and/or service ("in %" and/or "in h")')
            ]);

        $validator
            ->add('services', 'custom', [
                'rule'    => [$this, 'checkOutageDuration'],
                'message' => __('Outage duration needs to be numeric')
            ]);

        return $validator;
    }

    public function checkOutageDuration($context) {
        if (!is_array($context)) {
            return false;
        }

        foreach ($context as $record) {
            if (isset($record['outage']) && $record['outage'] !== null && $record['outage'] !== '') {
                return is_numeric($record['outage']);
            }
        }

        return true;
    }

    public function atLeastOneHostOrServiceInReport($context, $data) {
        if (!empty($data['data']['hosts'])) {
            foreach ($data['data']['hosts'] as $host) {
                if (isset($host['percent'])) {
                    if ((int)$host['percent'] === 1) {
                        return true;
                    }
                }

                if (isset($host['minute'])) {
                    if ((int)$host['minute'] === 1) {
                        return true;
                    }
                }
            }
        }

        if (!empty($data['data']['services'])) {
            foreach ($data['data']['services'] as $service) {
                if (isset($service['percent'])) {
                    if ((int)$service['percent'] === 1) {
                        return true;
                    }
                }

                if (isset($service['minute'])) {
                    if ((int)$service['minute'] === 1) {
                        return true;
                    }
                }
            }
        }

        return false;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param \Cake\ORM\RulesChecker $rules The rules object to be modified.
     * @return \Cake\ORM\RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        /*
        $rules->add($rules->existsIn(['host_id'], 'Hosts'));
        $rules->add($rules->existsIn(['service_id'], 'Services'));
*/
        return $rules;
    }

    /**
     * @param mixed $value
     * @param array $context
     * @return bool
     *
     * Custom validation rule for users if send interval not NEVER
     */
    public function atLeastOne($value, $context) {
        if (empty($context['data']['report_send_interval']) || $context['data']['report_send_interval'] === 'NEVER') {
            return true;
        }
        return !empty($context['data']['users']['_ids']);
    }


    /**
     * @param AutoreportFilter $AutoreportFilter
     * @param PaginateOMat|null $PaginateOMat
     * @param array $MY_RIGHTS
     * @return array
     */
    public function getAutoreportsIndex(AutoreportFilter $AutoreportFilter, PaginateOMat $PaginateOMat = null, $MY_RIGHTS = []) {
        $query = $this->find('all')
            ->select([
                'Autoreports.id',
                'Autoreports.name',
                'Autoreports.description',
                'Autoreports.container_id',
                'Autoreports.report_interval',
                'Autoreports.report_send_interval',
                'Autoreports.min_availability',
                'Autoreports.max_number_of_outages'
            ])
            ->group('Autoreports.id')
            ->disableHydration();

        $indexFilter = $AutoreportFilter->indexFilter();
        if (!empty($MY_RIGHTS)) {
            $indexFilter['Autoreports.container_id IN'] = $MY_RIGHTS;
        }
        $query->where($indexFilter);
        $query->order(
            array_merge(
                $AutoreportFilter->getOrderForPaginator('Autoreports.name', 'asc'),
                ['Autoreports.id' => 'asc']
            )
        );

        if ($PaginateOMat === null) {
            //Just execute query
            $result = $query->toArray();
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }

    /**
     * @param int $id
     * @return bool
     */
    public function existsById($id) {
        return $this->exists(['Autoreports.id' => $id]);
    }

    /**
     * @param int $id
     * @return bool
     */
    public function existsByIdForAvailabilityCalculation($id) {
        return $this->exists([
            'Autoreports.id'                       => $id,
            'Autoreports.min_availability_percent' => 1
        ]);
    }


    /**
     * @param int $id
     * @return array|Autoreport|null
     */
    public function getAutoreportById($id) {
        $query = $this->find()
            ->where([
                'Autoreports.id' => $id
            ])
            ->first();
        return $query;
    }


    /**
     * @param AutoreportServiceConditions $AutoreportServiceConditions
     * @param $MY_RIGHTS
     * @param array $selected
     * @return array|false|void
     */
    public function getAutoreportServicesForAngularCake4(AutoreportServiceConditions $AutoreportServiceConditions, $MY_RIGHTS, $selected = []) {
        /** @var $HostsTable HostsTable $query */
        $HostsTable = TableRegistry::getTableLocator()->get('Hosts');
        try {
            $query = $HostsTable->find()
                ->distinct('Hosts.id')
                ->select([
                    'Hosts.id',
                    'Hosts.name'
                ])
                ->innerJoinWith('HostsToContainersSharing', function (Query $q) use ($MY_RIGHTS) {
                    if (!empty($MY_RIGHTS)) {
                        return $q->where(['HostsToContainersSharing.id IN' => $MY_RIGHTS]);
                    }
                    return $q;
                })
                ->where([
                    'Hosts.id IN' => $AutoreportServiceConditions->getHostIds()
                ])
                ->contain([
                    'Services' => function (Query $q) use ($AutoreportServiceConditions, $selected) {

                        $q = $q->innerJoinWith('Servicetemplates')
                            ->select([
                                'Services.id',
                                'Services.host_id',
                                'servicename' => $q->newExpr('IF(Services.name IS NULL, Servicetemplates.name, Services.name)'),
                                'Services.disabled'
                            ]);
                        if (!empty($AutoreportServiceConditions->getServicenameRegex())) {
                            $q->having(['servicename REGEXP' => $AutoreportServiceConditions->getServicenameRegex()]);
                        }
                        if (!empty($selected)) {
                            $selected = array_filter($selected);
                            $q->where([
                                'Services.id NOT IN' => $selected
                            ]);
                        }
                        return $q;
                    }
                ]);

            $query->order('Hosts.name')
                ->disableHydration()
                ->all();

            $result = $this->emptyArrayIfNull($query->toArray());
            if (empty($selected)) {
                return $result;
            }
            $query = $HostsTable->find()
                ->distinct('Hosts.id')
                ->select([
                    'Hosts.id',
                    'Hosts.name'
                ])
                ->innerJoinWith('HostsToContainersSharing', function (Query $q) use ($MY_RIGHTS) {
                    if (!empty($MY_RIGHTS)) {
                        return $q->where(['HostsToContainersSharing.id IN' => $MY_RIGHTS]);
                    }
                    return $q;
                })
                ->where([
                    'Hosts.id IN' => $AutoreportServiceConditions->getHostIds()
                ])
                ->contain([
                    'Services' => function (Query $q) use ($AutoreportServiceConditions, $selected) {
                        $q = $q->innerJoinWith('Servicetemplates')
                            ->select([
                                'Services.id',
                                'Services.host_id',
                                'servicename' => $q->newExpr('IF(Services.name IS NULL, Servicetemplates.name, Services.name)'),
                                'Services.disabled'
                            ]);
                        if (!empty($selected)) {
                            $selected = array_filter($selected);
                            $q->where([
                                'Services.id IN' => $selected
                            ]);
                        }
                        return $q;
                    }
                ]);

            $query->order('Hosts.name')
                ->disableHydration()
                ->all();
            $selectedResult = $this->emptyArrayIfNull($query->toArray());
            foreach ($selectedResult as $selectedServices) {
                $hostname = $selectedServices['name'];
                foreach ($result as $key => $services) {
                    if ($services['name'] === $hostname) {
                        $result[$key]['services'] = array_merge($services['services'], $selectedServices['services']);
                    }

                }
            }

            return $result;
        } catch (\Exception $e) {
            return [];
        }
    }


    /**
     * @param int $id
     * @param array $requestData
     * @return array
     */
    public function prepareStepTwoDataForSave($id, $requestData) {
        $data = [
            'id'       => $id,
            'hosts'    => [],
            'services' => []
        ];

        $hostConfigurationOptions = [
            'percent'     => 1,
            'minute'      => 2,
            'alias'       => 4,
            'allfailures' => 8
        ];

        foreach ($requestData['hosts'] as $host) {
            $bitValue = 0;
            foreach ($hostConfigurationOptions as $key => $value) {
                if (isset($host[$key]) && (int)$host[$key] === 1) {
                    $bitValue += $value;
                }
            }
            if ($bitValue & 1 || $bitValue & 2) { //Is at least percent or minute checked?
                $data['hosts'][] = [
                    'id'        => $host['host_id'],
                    '_joinData' => [
                        'outage_duration'      => $host['outage'] ?? null,
                        'configuration_option' => $bitValue
                    ]
                ];
            }
        }

        $serviceConfigurationOptions = [
            'percent'     => 1,
            'minute'      => 2,
            'graph'       => 4,
            'allfailures' => 8
        ];

        foreach ($requestData['services'] as $service) {
            $bitValue = 0;
            foreach ($serviceConfigurationOptions as $key => $value) {
                if (isset($service[$key]) && (int)$service[$key] === 1) {
                    $bitValue += $value;
                }
            }

            if ($bitValue & 1 || $bitValue & 2) { //Is at least percent or minute checked?
                $data['services'][] = [
                    'id'        => $service['service_id'],
                    '_joinData' => [
                        'host_id'              => $service['host_id'],
                        'outage_duration'      => $service['outage'] ?? null,
                        'configuration_option' => $bitValue,
                        'graph_settings'       => $service['graphSettings'] ?? 'AVERAGE'
                    ]
                ];
            }
        }
        return $data;
    }

    /**
     * @param int $timeperiodId
     * @return bool
     */
    public function isTimeperiodUsedByAutoreports($timeperiodId) {
        $count = $this->find()
            ->where([
                'timeperiod_id' => $timeperiodId,
            ])->count();

        if ($count > 0) {
            return true;
        }

        return false;
    }

    /**
     * @param int $id
     * @return array|\AutoreportModule\Model\Entity\Autoreport
     */
    public function getAutoreportForOverview($id) {
        $autoreport = $this->find()
            ->where([
                'Autoreports.id' => $id
            ])
            ->contain([
                'Timeperiods' => [
                    'TimeperiodTimeranges'
                ],
                'Hosts',
                'Services'    => [
                    'Servicetemplates',
                    'Hosts'
                ],
                'Users',
                'Calendars'
            ])
            ->firstOrFail();

        $autoreport = $autoreport->toArray();

        $minAvailabilityString = '';
        if ($autoreport['min_availability'] !== null) {
            if ($autoreport['min_availability_percent']) {
                $minAvailabilityString = __('{0} %', $autoreport['min_availability']);
            } else {
                $minAvailabilityString = __('{0} minutes', $autoreport['min_availability']);
            }
        }
        $autoreport['min_availability_string'] = $minAvailabilityString;

        $considerHolidaysString = '';
        if ($autoreport['calendar_id'] > 0 && $autoreport['calendar']['name']) {
            $considerHolidaysString = $autoreport['calendar']['name'];
        }
        $autoreport['consider_holidays_string'] = $considerHolidaysString;

        $hostsWithServices = [];
        foreach ($autoreport['hosts'] as $host) {
            $hostsWithServices[$host['id']] = $host;
            $hostsWithServices[$host['id']]['services'] = [];
        }

        foreach ($autoreport['services'] as $service) {
            $service['servicename'] = $service['name'];
            if ($service['servicename'] === null || $service['servicename'] === '') {
                $service['servicename'] = $service['servicetemplate']['name'];
            }

            //Is the host missing in the report (because only services got selected?)
            if (!isset($hostsWithServices[$service['host_id']])) {
                $hostsWithServices[$service['host_id']] = $service['host'];
                $hostsWithServices[$service['host_id']]['services'] = [];
            }

            $hostsWithServices[$service['host_id']]['services'][] = $service;
        }

        $autoreport['hostsWithServices'] = array_values($hostsWithServices); //Remove host id as key to get an array

        $weekdays = [
            1 => __('Monday'),
            2 => __('Tuesday'),
            3 => __('Wednesday'),
            4 => __('Thursday'),
            5 => __('Friday'),
            6 => __('Saturday'),
            7 => __('Sunday'),
        ];
        foreach ($autoreport['timeperiod']['timeperiod_timeranges'] as $index => $timerange) {
            $autoreport['timeperiod']['timeperiod_timeranges'][$index]['weekday'] = $weekdays[$timerange['day']];
        }

        return $autoreport;
    }

    /**
     * @param int $id
     * @return array
     */
    public function getAutoreportForEditStepOne($id) {
        $autoreport = $this->find()
            ->where([
                'Autoreports.id' => $id
            ])
            ->contain([
                'Users'
            ])
            ->disableHydration()
            ->firstOrFail();

        $autoreport['users'] = [
            '_ids' => Hash::extract($autoreport, 'users.{n}.id')
        ];

        // Fix datatypes for AngularJS Frontend
        if (!empty($autoreport)) {
            $autoreport['show_time'] = (string)$autoreport['show_time'] ?? null;
            $autoreport['consider_downtimes'] = (int)$autoreport['consider_downtimes'] ?? 0;
            $autoreport['check_hard_state'] = (int)$autoreport['check_hard_state'] ?? 0;
        }


        return $autoreport;
    }

    /**
     * @param int $id
     * @return array
     */
    public function getAutoreportForEditStepTwo($id) {
        $autoreport = $this->find()
            ->where([
                'Autoreports.id' => $id
            ])
            ->contain([
                'Hosts',
                'Services' => [
                    'Servicetemplates',
                    'Hosts'
                ],
            ])
            ->disableHydration()
            ->firstOrFail();

        //Build up hostsWithServices for the table
        $hostsWithServices = [];
        foreach ($autoreport['hosts'] as $host) {
            $hostsWithServices[$host['id']] = $host;
            $hostsWithServices[$host['id']]['services'] = [];
        }

        foreach ($autoreport['services'] as $service) {
            $service['servicename'] = $service['name'];
            if ($service['servicename'] === null || $service['servicename'] === '') {
                $service['servicename'] = $service['servicetemplate']['name'];
            }

            //Is the host missing in the report (because only services got selected?)
            if (!isset($hostsWithServices[$service['host_id']])) {
                $hostsWithServices[$service['host_id']] = $service['host'];
                $hostsWithServices[$service['host_id']]['services'] = [];
            }

            $hostsWithServices[$service['host_id']]['services'][] = $service;
        }

        $autoreport['hostsWithServices'] = array_values($hostsWithServices); //Remove host id as key to get an array


        //Generate report configuration for the check boxes
        $postData = [
            'hosts'    => [],
            'services' => []
        ];
        foreach ($autoreport['hosts'] as $host) {
            $postData['hosts'][$host['id']] = [
                'host_id'     => $host['id'],
                'percent'     => ($host['_joinData']['configuration_option'] & 1) ? 1 : 0, //percent = 1
                'minute'      => ($host['_joinData']['configuration_option'] & 2) ? 1 : 0, //minute = 2
                'alias'       => ($host['_joinData']['configuration_option'] & 4) ? 1 : 0, //alias = 4
                'outage'      => $host['_joinData']['outage_duration'] ?? '',
                'allfailures' => ($host['_joinData']['configuration_option'] & 8) ? 1 : 0  //outage = 8
            ];
        }

        foreach ($autoreport['services'] as $service) {
            $postData['services'][$service['id']] = [
                'host_id'       => $service['host_id'],
                'service_id'    => $service['id'],
                'percent'       => ($service['_joinData']['configuration_option'] & 1) ? 1 : 0, //percent = 1
                'minute'        => ($service['_joinData']['configuration_option'] & 2) ? 1 : 0, //minute = 2
                'graph'         => ($service['_joinData']['configuration_option'] & 4) ? 1 : 0, //graph = 4
                'graphSettings' => $service['_joinData']['graph_settings'] ?? 'AVERAGE',
                'outage'        => $service['_joinData']['outage_duration'] ?? '',
                'allfailures'   => ($service['_joinData']['configuration_option'] & 8) ? 1 : 0  //outage = 8
            ];
        }

        $autoreport['POST'] = $postData;

        return $autoreport;
    }

    /**
     * @param $timeperiodId
     * @param array $MY_RIGHTS
     * @param bool $enableHydration
     * @return array
     */
    public function getAutoreportsByTimeperiodId($timeperiodId, $MY_RIGHTS = [], $enableHydration = true) {
        $query = $this->find()
            ->select([
                'Autoreports.id',
                'Autoreports.name'
            ])
            ->where([
                'Autoreports.timeperiod_id' => $timeperiodId
            ]);

        if (!empty($MY_RIGHTS)) {
            $query->where([
                'Autoreports.container_id IN' => $MY_RIGHTS
            ]);
        }
        $query->order('Autoreports.name', 'asc');
        $query->enableHydration($enableHydration);

        $result = $query->all();
        return $this->emptyArrayIfNull($result->toArray());
    }

    /**
     * @param int $reportId
     * @param array $newHostIds
     * @param array $oldHostIds //Hosts that got removed from current report
     * @param array $newServiceIds
     * @param array $oldServiceIds //Services that got removed from current report
     */
    public function saveUsageFlags($reportId, array $newHostIds, array $oldHostIds, array $newServiceIds, array $oldServiceIds) {
        /** @var HostsTable $HostsTable */
        $HostsTable = TableRegistry::getTableLocator()->get('Hosts');
        /** @var ServicesTable $ServicesTable */
        $ServicesTable = TableRegistry::getTableLocator()->get('Services');

        /** @var AutoreportsHostsMembershipsTable $AutoreportsHostsMembershipsTable */
        $AutoreportsHostsMembershipsTable = TableRegistry::getTableLocator()->get('AutoreportModule.AutoreportsHostsMemberships');

        /** @var AutoreportsServicesMembershipsTable $AutoreportsServicesMembershipsTable */
        $AutoreportsServicesMembershipsTable = TableRegistry::getTableLocator()->get('AutoreportModule.AutoreportsServicesMemberships');

        foreach ($newHostIds as $newHostId) {
            $HostsTable->setUsageFlagById($newHostId, AUTOREPORT_MODULE);
        }

        foreach ($oldHostIds as $oldHostId) {
            //Check if host is used in other reports before removing the usage flag
            if ($AutoreportsHostsMembershipsTable->isHostUsedByReport($reportId, $oldHostId) === 0) {
                $HostsTable->removeUsageFlagById($oldHostId, AUTOREPORT_MODULE);
            }
        }

        foreach ($newServiceIds as $newServiceId) {
            $ServicesTable->setUsageFlagById($newServiceId, AUTOREPORT_MODULE);
        }
        foreach ($oldServiceIds as $oldServiceId) {
            //Check if service is used in other reports before removing the usage flag
            //Check if host is used in other reports before removing the usage flag
            if ($AutoreportsServicesMembershipsTable->isServiceUsedByReport($reportId, $oldServiceId) === 0) {
                $ServicesTable->removeUsageFlagById($oldServiceId, AUTOREPORT_MODULE);
            }
        }
    }

    /**
     * @param int|array $ids
     * @param array $ids
     * @param bool $enableHydration
     * @return array
     */
    public function getAutoreportsByIds($ids, $MY_RIGHTS, $enableHydration = true) {
        if (!is_array($ids)) {
            $ids = [$ids];
        }

        $query = $this->find()
            ->where([
                'Autoreports.id IN'           => $ids,
                'Autoreports.container_id IN' => $MY_RIGHTS
            ])
            ->order([
                'Autoreports.name' => 'asc'
            ])
            ->enableHydration($enableHydration)
            ->all();
        return $query->toArray();
    }

    /**
     * @param $id
     * @return array|\Cake\Datasource\EntityInterface
     */
    public function getAutoreportByIdForReporting($id, int $start, int $end, $enableHydration = false) {
        $startDateSqlFormat = date('Y-m-d H:i:s', $start);
        $endDateSqlFormat = date('Y-m-d H:i:s', $end);

        $query = $this->find()
            ->where([
                'Autoreports.id' => $id
            ])
            ->contain([
                'Hosts'       => function (Query $query) {
                    $query
                        ->disableAutoFields()
                        ->select([
                            'Hosts.id',
                            'Hosts.name',
                            'Hosts.uuid',
                            'Hosts.description',
                            'Hosts.created'
                        ]);
                    return $query;
                },
                'Services'    => function (Query $query) {
                    $query
                        ->disableAutoFields()
                        ->select([
                            'Services.id',
                            'Services.name',
                            'Services.uuid',
                            'Services.host_id',
                            'Services.servicetemplate_id',
                            'Services.description',
                            'Services.created',
                            'Services.service_type'
                        ])
                        ->contain([
                            'Hosts'            => function (Query $query) {
                                $query
                                    ->disableAutoFields()
                                    ->select([
                                        'Hosts.id',
                                        'Hosts.name',
                                        'Hosts.uuid',
                                        'Hosts.description',
                                        'Hosts.created'
                                    ]);
                                return $query;
                            },
                            'Servicetemplates' => function (Query $query) {
                                $query
                                    ->disableAutoFields()
                                    ->select([
                                        'Servicetemplates.id',
                                        'Servicetemplates.name',
                                        'Servicetemplates.uuid',
                                        'Servicetemplates.description',
                                        'Servicetemplates.created'
                                    ]);
                                return $query;
                            }
                        ]);
                    return $query;
                },
                'Timeperiods' => [
                    'TimeperiodTimeranges'
                ],
                'Calendars'   => [
                    'CalendarHolidays' => function (Query $query) use ($startDateSqlFormat, $endDateSqlFormat) {
                        $query->where([
                            'CalendarHolidays.date BETWEEN :start AND :end'
                        ])
                            ->bind(':start', $startDateSqlFormat, 'date')
                            ->bind(':end', $endDateSqlFormat, 'date');
                        return $query;
                    }
                ]
            ])
            ->enableHydration($enableHydration);

        return $query->firstOrFail();
    }

    /**
     * @param $id
     * @return array|\Cake\Datasource\EntityInterface
     */
    public function getAutoreportByIdForAvailabilityWidget($id) {

        $query = $this->find()
            ->where([
                'Autoreports.id'                       => $id,
                'Autoreports.min_availability_percent' => 1
            ])
            ->contain([
                'AutoreportAvailabilityLog' => function (Query $query) {
                    $query
                        ->disableAutoFields()
                        ->select([
                            'AutoreportAvailabilityLog.id',
                            'AutoreportAvailabilityLog.autoreport_id',
                            'AutoreportAvailabilityLog.last_update',
                            'AutoreportAvailabilityLog.minimal_availability_percent',
                            'AutoreportAvailabilityLog.availability_total_time',
                            'AutoreportAvailabilityLog.determined_availability_time',
                            'AutoreportAvailabilityLog.determined_availability_percent'
                        ])
                        ->order(['AutoreportAvailabilityLog.last_update' => 'desc'])
                        ->limit(20);
                    return $query;
                }
            ])
            ->disableHydration();
        $result = $query->first();
        if (empty($result)) {
            return [];
        }
        return $result;
    }

    /**
     * @param array $containerIds
     * @param bool $availabilityLog
     * @return array
     */
    public function getAutoreportsByContainerIdsAsList($containerIds = [], $availabilityLog = false) {
        if (!is_array($containerIds)) {
            $containerIds = [$containerIds];
        }
        /** @var AutoreportsTable $query */
        $query = $this->find()
            ->select([
                'Autoreports.id',
                'Autoreports.name'
            ])
            ->where(['Autoreports.container_id IN' => $containerIds]);

        if ($availabilityLog === true) {
            $query->andWhere([
                'Autoreports.min_availability_percent' => 1
            ]);
        }

        $autoreports = $query->all();

        if (empty($autoreports)) {
            return [];
        }
        return $autoreports->toArray();
    }

    /**
     * @return array
     */
    public function getAutoreportsForCronjob() {
        /** @var AutoreportsTable $query */
        $query = $this->find()
            ->contain([
                'Containers',
                'Users'
            ])
            ->select([
                'Autoreports.id',
                'Autoreports.container_id',
                'Autoreports.name',
                'Autoreports.report_send_interval',
                'Autoreports.last_send_date',
                'Autoreports.use_start_time',
                'Autoreports.report_start_date',
                'Autoreports.min_availability_percent',
                'Autoreports.min_availability',
                'Autoreports.last_number_of_outages',
                'Containers.name'
            ]);

        $autoreports = $query->all();
        if (empty($autoreports)) {
            return [];
        }
        return $autoreports->toArray();
    }

    /**
     * @return array
     */
    public function getAutoreportsAsListForCalculateAvailabilityCronjob() {
        /** @var AutoreportsTable $query */
        $query = $this->find()
            ->select([
                'Autoreports.id',
                'Autoreports.name',
                'Autoreports.report_interval'
            ])
            ->where(['Autoreports.min_availability_percent' => 1]);

        $autoreports = $query->all();
        if (empty($autoreports)) {
            return [];
        }
        return $autoreports->toArray();
    }

    /**
     * @param int $containerId
     * @param string $type
     * @param array $MY_RIGHTS
     * @param array $where
     * @return array
     */
    public function getAutoreportsByContainerIdExact($containerId, $type = 'all', $index = 'id', $MY_RIGHTS = [], $where = []) {
        $_where = [
            'Autoreports.container_id' => $containerId
        ];

        $where = Hash::merge($_where, $where);

        $query = $this->find();
        $query->select([
            'Autoreports.' . $index,
            'Autoreports.name'
        ]);
        $query->where($where);

        if (!empty($MY_RIGHTS)) {
            $query->andWhere([
                'Autoreports.container_id IN' => $MY_RIGHTS
            ]);
        }

        $query->disableHydration();

        $result = $query->toArray();
        if (empty($result)) {
            return [];
        }

        if ($type === 'all') {
            return $result;
        }

        $list = [];
        foreach ($result as $row) {
            $list[$row[$index]] = $row['name'];
        }

        return $list;
    }

    /**
     * @param $id
     * @return array|\Cake\Datasource\EntityInterface
     */
    public function getAutoreportForDelete($id) {
        return $this->find()
            ->contain([
                'Hosts',
                'Services'
            ])
            ->where([
                'Autoreports.id' => $id
            ])
            ->firstOrFail();
    }

}
